
/******************************************************************************
* MODULE     : base_types.cc
* DESCRIPTION: Basic types for the Mathemagix interpreter
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <extern.hh>
#include <single.hh>
#include <literal.hh>
#include <printer.hh>
#include <serializer.hh>
#include <scheme.hh>
#include <pretty.hh>
#include <simplify.hh>
#include <ccpp.hh>
namespace mmx {

/******************************************************************************
* Booleans
******************************************************************************/

generic
bool_new (generic g) {
  (void) g;
  return single (false);
}

generic
bool_copy (generic g) {
  return single (as_bool (g[1]));
}

generic
bool_hash (generic g) {
  return g[1];
}

generic
bool_not (generic g) {
  return single (!as_bool (g[1]));
}

generic
bool_or (generic g) {
  return single (as_bool (g[1]) | as_bool (g[2]));
}

generic
bool_xor (generic g) {
  return single (as_bool (g[1]) ^ as_bool (g[2]));
}

generic
bool_and (generic g) {
  return single (as_bool (g[1]) & as_bool (g[2]));
}

generic
bool_implies (generic g) {
  return single (as_bool (g[2]) | (!as_bool (g[2])));
}

generic
bool_equivalent (generic g) {
  return single (as_bool (g[1]) == as_bool (g[2]));
}

generic
bool_equal (generic g) {
  return single (as_bool (g[1]) == as_bool (g[2]));
}

generic
bool_not_equal (generic g) {
  return single (as_bool (g[1]) != as_bool (g[2]));
}

generic
bool_output (generic g) {
  printer mmx_out= as_printer (g[1]);
  ostream& out (mmx_out->out);
  if (as_bool (g[2])) out << "true"; else out << "false";
  out.flush();
  return mmx_out;
}

/******************************************************************************
* Integers
******************************************************************************/

generic
int_new (generic g) {
  (void) g;
  return single (0);
}

generic
int_copy (generic g) {
  return single (as_int (g[1]));
}

generic
int_literal (generic g) {
  symbol s= S (as_literal (g[1]));
  return single (as_int (s));
}

generic
int_hash (generic g) {
  return g[1];
}

generic
int_plus (generic g) {
  return single (as_int (g[1]) + as_int (g[2]));
}

generic
int_neg (generic g) {
  return single (-as_int (g[1]));
}

generic
int_minus (generic g) {
  return single (as_int (g[1]) - as_int (g[2]));
}

generic
int_times (generic g) {
  return single (as_int (g[1]) * as_int (g[2]));
}

generic
int_div (generic g) {
  return single (as_int (g[1]) / as_int (g[2]));
}

generic
int_mod (generic g) {
  return single (as_int (g[1]) % as_int (g[2]));
}

generic
int_equal (generic g) {
  return single (as_int (g[1]) == as_int (g[2]));
}

generic
int_not_equal (generic g) {
  return single (as_int (g[1]) != as_int (g[2]));
}

generic
int_less (generic g) {
  return single (as_int (g[1]) < as_int (g[2]));
}

generic
int_less_equal (generic g) {
  return single (as_int (g[1]) <= as_int (g[2]));
}

generic
int_greater (generic g) {
  return single (as_int (g[1]) > as_int (g[2]));
}

generic
int_greater_equal (generic g) {
  return single (as_int (g[1]) >= as_int (g[2]));
}

generic
int_output (generic g) {
  printer mmx_out= as_printer (g[1]);
  ostream& out (mmx_out->out);
  out << as_int (g[2]);
  out.flush();
  return mmx_out;
}

/******************************************************************************
* Strings
******************************************************************************/

generic
string_new (generic g) {
  (void) g;
  return string ("");
}

generic
string_copy (generic g) {
  return string (as_string (g[1]));
}

generic
string_literal (generic g) {
  symbol s= S (as_literal (g[1]));
  return unquote (as_string (S (s)));
}

generic
string_hash (generic g) {
  return single (hash (as_string (g[1])));
}

generic
string_concat (generic g) {
  return string (as_string (g[1]) * as_string (g[2]));
}

generic
string_equal (generic g) {
  return single (as_string (g[1]) == as_string (g[2]));
}

generic
string_not_equal (generic g) {
  return single (as_string (g[1]) != as_string (g[2]));
}

generic
string_less (generic g) {
  return single (as_string (g[1]) < as_string (g[2]));
}

generic
string_less_equal (generic g) {
  return single (as_string (g[1]) <= as_string (g[2]));
}

generic
string_greater (generic g) {
  return single (as_string (g[2]) < as_string (g[1]));
}

generic
string_greater_equal (generic g) {
  return single (as_string (g[2]) <= as_string (g[1]));
}

generic
string_pretty (generic g) {
  string  s= as_string (g[1]);
  generic r= string_to_scheme (s);
  r= simplify (r);
  serializer sout;
  pretty_print (sout->out, r);
  // cout << lf << lf << "sout= " << sout << lf << lf;
  return as_string (sout);
}

generic
string_ccpp (generic g) {
  string  s= as_string (g[1]);
  generic r= string_to_scheme (s);
  serializer sout;
  ccpp_print (sout->out, r);
  return as_string (sout);
}

generic
string_quote (generic g) {
  return quote (as_string (g[1]));
}

generic
string_unquote (generic g) {
  return unquote (as_string (g[1]));
}

generic
string_output_raw (generic g) {
  printer mmx_out= as_printer (g[1]);
  ostream& out (mmx_out->out);
  string s= as_string (g[2]);
  out << delsimpl (s);
  out.flush();
  return mmx_out;
}

generic
string_output (generic g) {
  printer mmx_out= as_printer (g[1]);
  ostream& out (mmx_out->out);
  string s= as_string (g[2]);
  out << quote (s);
  out.flush();
  return mmx_out;
}

/******************************************************************************
* Output string streams
******************************************************************************/

generic
void_to_ostrstream (generic g) {
  (void) g;
  return serializer ();
}

generic
ostrstream_to_string (generic g) {
  return as_string (as_serializer (g[1]));
}

generic
ostrstream_to_ostream (generic g) {
  return as_printer (as_serializer (g[1]));
}

/******************************************************************************
* Interface
******************************************************************************/

void
initialize_base_types () {
  generic ID_VOID_BOOL= function_type (ID_VOID, ID_BOOL);
  generic ID_BOOL_INT= function_type (ID_BOOL, ID_INT);
  generic ID_BOOL_BOOL= function_type (ID_BOOL, ID_BOOL);
  generic ID_BOOL_BOOL_BOOL= function_type (ID_BOOL, ID_BOOL, ID_BOOL);
  generic ID_LITERAL_INT_INT= function_type (ID_LITERAL_INT, ID_INT);
  generic ID_VOID_INT= function_type (ID_VOID, ID_INT);
  generic ID_INT_INT= function_type (ID_INT, ID_INT);
  generic ID_INT_INT_INT= function_type (ID_INT, ID_INT, ID_INT);
  generic ID_INT_INT_BOOL= function_type (ID_INT, ID_INT, ID_BOOL);
  generic ID_LITERAL_STRING_STRING=
    function_type (ID_LITERAL_STRING, ID_STRING);
  generic ID_VOID_STRING= function_type (ID_VOID, ID_STRING);
  generic ID_STRING_INT= function_type (ID_STRING, ID_INT);
  generic ID_STRING_STRING= function_type (ID_STRING, ID_STRING);
  generic ID_STRING_STRING_STRING=
    function_type (ID_STRING, ID_STRING, ID_STRING);
  generic ID_STRING_STRING_BOOL= function_type (ID_STRING, ID_STRING, ID_BOOL);
  generic ID_VOID_OSTRSTREAM= function_type (ID_OSTRSTREAM);
  generic ID_OSTRSTREAM_STRING= function_type (ID_OSTRSTREAM, ID_STRING);
  generic ID_OSTRSTREAM_OSTREAM= function_type (ID_OSTRSTREAM, ID_OSTREAM);

  declare (MMX_CLASS, ID_CLASS, ID_CLASS);
  declare (MMX_GENERIC, ID_GENERIC, ID_CLASS);
  declare (MMX_VOID, ID_VOID, ID_CLASS);

  declare (MMX_BOOL, ID_BOOL, ID_CLASS);
  declare ("bool", bool_new, ID_VOID_BOOL);
  declare ("bool", bool_copy, ID_BOOL_BOOL);
  declare ("true", single (true), ID_BOOL);
  declare ("false", single (false), ID_BOOL);
  declare ("hash", bool_hash, ID_BOOL_INT);
  declare (MMX_NOT, bool_not, ID_BOOL_BOOL);
  declare (MMX_OR, bool_or, ID_BOOL_BOOL_BOOL);
  declare (MMX_XOR, bool_xor, ID_BOOL_BOOL_BOOL);
  declare (MMX_AND, bool_and, ID_BOOL_BOOL_BOOL);
  declare (MMX_IMPLIES, bool_implies, ID_BOOL_BOOL_BOOL);
  declare (MMX_EQUIV, bool_equivalent, ID_BOOL_BOOL_BOOL);
  declare (MMX_EQUAL, bool_equal, ID_BOOL_BOOL_BOOL);
  declare (MMX_UNEQUAL, bool_not_equal, ID_BOOL_BOOL_BOOL);
  declare (MMX_LESSLESS_STR, bool_output,
	   function_type (ID_OSTREAM, ID_BOOL, ID_OSTREAM));

  declare (MMX_INT, ID_INT, ID_CLASS);
  declare (MMX_LITERAL_INT, ID_LITERAL_INT, ID_CLASS);
  declare ("int", int_new, ID_VOID_INT);
  declare ("int", int_copy, ID_INT_INT);
  declare (MMX_NEW_INT, int_literal, ID_LITERAL_INT_INT);
  declare ("hash", int_hash, ID_INT_INT);
  declare (MMX_PLUS, int_plus, ID_INT_INT_INT);
  declare (MMX_MINUS, int_neg, ID_INT_INT);
  declare (MMX_MINUS, int_minus, ID_INT_INT_INT);
  declare (MMX_TIMES, int_times, ID_INT_INT_INT);
  declare (MMX_DIV, int_div, ID_INT_INT_INT);
  declare (MMX_MOD, int_mod, ID_INT_INT_INT);
  declare (MMX_EQUAL, int_equal, ID_INT_INT_BOOL);
  declare (MMX_UNEQUAL, int_not_equal, ID_INT_INT_BOOL);
  declare (MMX_LESS, int_less, ID_INT_INT_BOOL);
  declare (MMX_LESSEQ, int_less_equal, ID_INT_INT_BOOL);
  declare (MMX_GTR, int_greater, ID_INT_INT_BOOL);
  declare (MMX_GTREQ, int_greater_equal, ID_INT_INT_BOOL);
  declare (MMX_LESSLESS_STR, int_output,
	   function_type (ID_OSTREAM, ID_INT, ID_OSTREAM));

  declare (MMX_STRING, ID_STRING, ID_CLASS);
  declare ("string", string_new, ID_VOID_STRING);
  declare ("string", string_copy, ID_STRING_STRING);
  declare (MMX_LITERAL_STRING, ID_LITERAL_STRING, ID_CLASS);
  declare (MMX_NEW_STRING, string_literal, ID_LITERAL_STRING_STRING);
  declare ("hash", string_hash, ID_STRING_INT);
  declare (MMX_TIMES, string_concat, ID_STRING_STRING_STRING);
  declare (MMX_EQUAL, string_equal, ID_STRING_STRING_BOOL);
  declare (MMX_UNEQUAL, string_not_equal, ID_STRING_STRING_BOOL);
  declare (MMX_LESS, string_less, ID_STRING_STRING_BOOL);
  declare (MMX_LESSEQ, string_less_equal, ID_STRING_STRING_BOOL);
  declare (MMX_GTR, string_greater, ID_STRING_STRING_BOOL);
  declare (MMX_GTREQ, string_greater_equal, ID_STRING_STRING_BOOL);
  declare ("pretty", string_pretty, function_type (ID_STRING, ID_STRING));
  declare ("ccpp", string_ccpp, function_type (ID_STRING, ID_STRING));
  declare (MMX_LESSLESS, string_output_raw,
	   function_type (ID_OSTREAM, ID_STRING, ID_OSTREAM));
  declare (MMX_LESSLESS_VAR, string_output,
	   function_type (ID_OSTREAM, ID_STRING, ID_OSTREAM));
  declare (MMX_LESSLESS_STR, string_output,
	   function_type (ID_OSTREAM, ID_STRING, ID_OSTREAM));

  declare (MMX_OSTREAM, ID_OSTREAM, ID_CLASS);
  declare ("mmout", printer (symbol ("mmout"), cout), ID_OSTREAM);
  declare ("mmerr", printer (symbol ("mmerr"), cerr), ID_OSTREAM);

  declare (MMX_OSTRSTREAM, ID_OSTRSTREAM, ID_CLASS);
  declare ("ostrstream", void_to_ostrstream, ID_VOID_OSTRSTREAM);
  declare ("string", ostrstream_to_string, ID_OSTRSTREAM_STRING);
  declare ("ostream", ostrstream_to_ostream, ID_OSTRSTREAM_OSTREAM);
}

} // namespace mmx
