
/******************************************************************************
* MODULE     : declare.cc
* DESCRIPTION: Declaring new global or local variables
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <evaluator.hh>
namespace mmx {

static void
overload (list lv, list lt, list& rv, list& rt, value v) {
  if (nil (lv)) {
    rv= list (v->val );
    rt= list (v->type);
  }
  else if (car (lt) == v->type) {
    rv= cons (v->val , cdr (lv));
    rt= cons (v->type, cdr (lt));
  }
  else {
    overload (cdr (lv), cdr (lt), rv, rt, v);
    rv= cons (car (lv), rv);
    rt= cons (car (lt), rt);
  }
}

value
overload (value v1, value v2) {
  if (v1->type == v2->type) return v2;
  if (!is_func (v1->type, ID_AND))
    return value (list (MMX_AND, v2->val, v1->val),
		  list (ID_AND, v2->type, v1->type));
  else {
    list lv= cdr (as_list (v1->val )), rv;
    list lt= cdr (as_list (v1->type)), rt;
    overload (lv, lt, rv, rt, v2);
    return value (cons (MMX_AND, rv), cons (ID_AND, rt));
  }
}

void
evaluator_rep::declare (symbol var, value val) {
  // declare the symbol var as val; may also overload var

  if (env->local_contains (var)) {
    if (!is_value (env->local_read (var))) {
      error (var, "Invalid overloading");
      return;
    }
    env->local_write (var)= overload (as_value (env->local_read (var)), val);
  }
  else env->local_write (var)= val;
}

void
evaluator_rep::global_declare (symbol var, value val) {
  if ((var == MMX_NEW_INT) ||
      (var == MMX_NEW_FLOAT) ||
      (var == MMX_NEW_STRING))
    env->local_write (var)= val;
  else declare (var, val);
}

} // namespace mmx
