
/******************************************************************************
* MODULE     : initialize.cc
* DESCRIPTION: Initialize the standard routines
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <evaluator.hh>
#include <function.hh>
#include <single.hh>
#include <printer.hh>
#include <keyword.hh>
namespace mmx {

#define KW(kw,name,r) GC_GLOBAL(keyword,kw,keyword(name,r))

extern generic cycle_witness;
extern int cur_location;
extern list sources;
extern table source_location;

/******************************************************************************
* All keywords (multiple versions due to continuations)
* should be in same file as where the evaluator_rep constructor
******************************************************************************/

KW (KW_COMPOUND_1, "compound!", &evaluator_rep::kw_compound_1);
KW (KW_COMPOUND_2, "compound!2", &evaluator_rep::kw_compound_2);
KW (KW_COMPOUND_3, "compound!3", &evaluator_rep::kw_compound_3);
KW (KW_EVAL_ARGS_1, "eval_args!", &evaluator_rep::kw_eval_args_1);
KW (KW_LOAD_1, MMX_LOAD, &evaluator_rep::kw_load_1);
KW (KW_LOAD_2, "load!2", &evaluator_rep::kw_load_2);
KW (KW_SUPPORTS_1, MMX_SUPPORTS, &evaluator_rep::kw_supports_1);
KW (KW_SUPPORTS_2, "supports!2", &evaluator_rep::kw_supports_2);
KW (KW_USE_1, MMX_USE, &evaluator_rep::kw_use_1);
KW (KW_USE_2, "use!2", &evaluator_rep::kw_use_2);

KW (KW_CONV_ALIASES_1, CONV_ALIASES, &evaluator_rep::kw_conv_aliases_1);
KW (KW_CONV_ALIASES_2, "aliases#2", &evaluator_rep::kw_conv_aliases_2);
KW (KW_CONV_AMBIGUOUS_1, CONV_AMBIGUOUS, &evaluator_rep::kw_conv_ambiguous_1);
KW (KW_CONV_APPLY_1, CONV_APPLY, &evaluator_rep::kw_conv_apply_1);
KW (KW_CONV_APPLY_2, "apply#2", &evaluator_rep::kw_conv_apply_2);
KW (KW_CONV_APPLY_3, "apply#3", &evaluator_rep::kw_conv_apply_3);
KW (KW_CONV_CROSS_1, CONV_CROSS, &evaluator_rep::kw_conv_cross_1);
KW (KW_CONV_CROSS_2, "cross#2", &evaluator_rep::kw_conv_cross_2);
KW (KW_CONV_FORALL_1, CONV_FORALL, &evaluator_rep::kw_conv_forall_1);
KW (KW_CONV_IGNORE_1, CONV_IGNORE, &evaluator_rep::kw_conv_ignore_1);
KW (KW_CONV_LAMBDA_1, CONV_LAMBDA, &evaluator_rep::kw_conv_lambda_1);
KW (KW_CONV_SINGLET_1, CONV_SINGLET, &evaluator_rep::kw_conv_singlet_1);
KW (KW_CONV_SINGLET_2, "singlet#2", &evaluator_rep::kw_conv_singlet_2);
KW (KW_CONV_TEMPLATE_1, CONV_TEMPLATE, &evaluator_rep::kw_conv_template_1);
KW (KW_CONV_TEMPLATE_2, "template#2", &evaluator_rep::kw_conv_template_2);
KW (KW_CONV_TEMPLATE_3, "template#3", &evaluator_rep::kw_conv_template_3);
KW (KW_CONV_UNALIAS_1, CONV_UNALIAS, &evaluator_rep::kw_conv_unalias_1);
KW (KW_CONV_UNALIAS_2, "unalias#2", &evaluator_rep::kw_conv_unalias_2);
KW (KW_CONV_UNAND_1, CONV_UNAND, &evaluator_rep::kw_conv_unand_1);
KW (KW_CONV_UNAND_2, "unand#2", &evaluator_rep::kw_conv_unand_2);
KW (KW_CONV_UNAND_3, "unand#3", &evaluator_rep::kw_conv_unand_3);
KW (KW_CONV_UNCROSS_1, CONV_UNCROSS, &evaluator_rep::kw_conv_uncross_1);
KW (KW_CONV_UNCROSS_2, "uncross#2", &evaluator_rep::kw_conv_uncross_2);
KW (KW_CONV_VCROSS_1, CONV_VCROSS, &evaluator_rep::kw_conv_vcross_1);
KW (KW_CONV_VCROSS_2, "vcross#2", &evaluator_rep::kw_conv_vcross_2);

KW (KW_ASSIGN_1, MMX_ASSIGN, &evaluator_rep::kw_assign_1);
KW (KW_ASSIGN_2, ":=!2", &evaluator_rep::kw_assign_2);
KW (KW_ASSIGN_3, ":=!3", &evaluator_rep::kw_assign_3);
KW (KW_BEGIN_1, MMX_BEGIN, &evaluator_rep::kw_begin_1);
KW (KW_BEGIN_2, "begin!2", &evaluator_rep::kw_begin_2);
KW (KW_BREAK_1, MMX_BREAK, &evaluator_rep::kw_break_1);
KW (KW_CONTINUE_1, MMX_CONTINUE, &evaluator_rep::kw_continue_1);
KW (KW_CROSS_1, MMX_CROSS, &evaluator_rep::kw_cross_1);
KW (KW_CROSS_2, "Cross!2", &evaluator_rep::kw_cross_2);
KW (KW_CASE_1, MMX_CASE, &evaluator_rep::kw_case_1);
KW (KW_CASE_2, "case!2", &evaluator_rep::kw_case_2);
KW (KW_CASE_3, "case!3", &evaluator_rep::kw_case_3);
KW (KW_CASE_4, "case!4", &evaluator_rep::kw_case_4);
KW (KW_DEFINEALIAS_1, MMX_DEFINEALIAS, &evaluator_rep::kw_define_1);
KW (KW_DEFINE_1, MMX_DEFINE, &evaluator_rep::kw_define_1);
KW (KW_DEFINE_2, "==!2", &evaluator_rep::kw_define_2);
KW (KW_DEFINE_3, "==!3", &evaluator_rep::kw_define_3);
KW (KW_EXISTS_1, MMX_EXISTS, &evaluator_rep::kw_quantor_1);
KW (KW_FOR_1, MMX_FOR, &evaluator_rep::kw_for_1);
KW (KW_FORALL_1, MMX_FORALL, &evaluator_rep::kw_quantor_1);
KW (KW_FORALLVAL_1, MMX_FORALLVAL, &evaluator_rep::kw_forallval_1);
KW (KW_FORALLVAL_2, "forallval!2", &evaluator_rep::kw_forallval_2);
KW (KW_FORIN_1, MMX_FORIN, &evaluator_rep::kw_forin_1);
KW (KW_FORIN_2, "forin!2", &evaluator_rep::kw_forin_2);
KW (KW_FORIN_3, "forin!3", &evaluator_rep::kw_forin_3);
KW (KW_FORIN_4, "forin!4", &evaluator_rep::kw_forin_4);
KW (KW_FORIN_5, "forin!5", &evaluator_rep::kw_forin_5);
KW (KW_FORIN_6, "forin!6", &evaluator_rep::kw_forin_6);
KW (KW_FUNCTION_1, MMX_FUNCTION, &evaluator_rep::kw_function_1);
KW (KW_FUNCTION_2, "->!2", &evaluator_rep::kw_function_2);
KW (KW_GENERATE_1, MMX_GENERATE, &evaluator_rep::kw_generate_1);
KW (KW_GENERATE_2, "generate!2", &evaluator_rep::kw_generate_2);
KW (KW_GENERATE_3, "generate!3", &evaluator_rep::kw_generate_3);
KW (KW_IF_1, MMX_IF, &evaluator_rep::kw_if_1);
KW (KW_IF_2, "if!2", &evaluator_rep::kw_if_2);
KW (KW_JOIN_1, MMX_JOIN, &evaluator_rep::kw_join_1);
KW (KW_JOIN_2, "join!2", &evaluator_rep::kw_join_2);
KW (KW_JOIN_3, "join!3", &evaluator_rep::kw_join_3);
KW (KW_LAMBDA_1, MMX_LAMBDA, &evaluator_rep::kw_lambda_1);
KW (KW_LAMBDA_2, "lambda!2", &evaluator_rep::kw_lambda_2);
KW (KW_LOCAL_1, MMX_LOCAL, &evaluator_rep::kw_local_1);
KW (KW_LOCAL_2, "local!2", &evaluator_rep::kw_local_2);
KW (KW_LOOP_1, MMX_LOOP, &evaluator_rep::kw_loop_1);
KW (KW_LOOP_2, "loop!2", &evaluator_rep::kw_loop_2);
KW (KW_LOOP_3, "loop!3", &evaluator_rep::kw_loop_3);
KW (KW_LOOP_4, "loop!4", &evaluator_rep::kw_loop_4);
KW (KW_QUANTOR_2, "quantor!2", &evaluator_rep::kw_quantor_2);
KW (KW_RETRIEVE_1, MMX_RETRIEVE, &evaluator_rep::kw_retrieve_1);
KW (KW_RETRIEVE_2, "retrieve!2", &evaluator_rep::kw_retrieve_2);
KW (KW_RETRIEVE_3, "retrieve!3", &evaluator_rep::kw_retrieve_3);
KW (KW_RETURN_1, MMX_RETURN, &evaluator_rep::kw_return_1);
KW (KW_RETURN_2, "return!2", &evaluator_rep::kw_return_2);
KW (KW_SPLIT_1, MMX_SPLIT, &evaluator_rep::kw_split_1);
KW (KW_SPLIT_2, "split!2", &evaluator_rep::kw_split_2);
KW (KW_STEP_1, MMX_STEP, &evaluator_rep::kw_step_1);
KW (KW_TRANSTYPE_1, MMX_TRANSTYPE, &evaluator_rep::kw_transtype_1);
KW (KW_TRANSTYPE_2, ":>!2", &evaluator_rep::kw_transtype_2);
KW (KW_TUPLE_1, MMX_TUPLE, &evaluator_rep::kw_tuple_1);
KW (KW_TUPLE_2, "tuple!2", &evaluator_rep::kw_tuple_2);
KW (KW_UNTIL_1, MMX_UNTIL, &evaluator_rep::kw_until_1);
KW (KW_VCROSS_1, MMX_VCROSS, &evaluator_rep::kw_cross_1);
KW (KW_VCROSS_2, "VCross!2", &evaluator_rep::kw_cross_2);
KW (KW_VTUPLE_1, MMX_VTUPLE, &evaluator_rep::kw_tuple_1);
KW (KW_VTUPLE_2, "vtuple!2", &evaluator_rep::kw_tuple_2);
KW (KW_WHILE_1, MMX_WHILE, &evaluator_rep::kw_while_1);
KW (KW_YIELD_1, MMX_YIELD, &evaluator_rep::kw_yield_1);
KW (KW_YIELD_2, "yield!2", &evaluator_rep::kw_yield_2);

GC_GLOBAL (table, mmx_lib_name, table ());
GC_GLOBAL (symbol, MMX_PRINT_ANY, "print_any!");
GC_GLOBAL (symbol, MMX_PRINT_CLASS, "print_class!");
GC_GLOBAL (symbol, MMX_PRINT_CROSS, "print_cross!");
GC_GLOBAL (symbol, MMX_PRINT_TUPLE, "print_tuple!");
GC_GLOBAL (symbol, MMX_PRINT_ALIAS, "print_forall!");
GC_GLOBAL (symbol, MMX_PRINT_CODE, "print_code!");

KW (LIB_PRINT_ANY_1, MMX_PRINT_ANY, &evaluator_rep::lib_print_any_1);
KW (LIB_PRINT_CLASS_1, MMX_PRINT_CLASS, &evaluator_rep::lib_print_class_1);
KW (LIB_PRINT_CLASS_2, "print_class!2", &evaluator_rep::lib_print_class_2);
KW (LIB_PRINT_CLASS_3, "print_class!3", &evaluator_rep::lib_print_class_3);
KW (LIB_PRINT_CROSS_1, MMX_PRINT_CROSS, &evaluator_rep::lib_print_cross_1);
KW (LIB_PRINT_CROSS_2, "print_cross!2", &evaluator_rep::lib_print_cross_2);
KW (LIB_PRINT_CROSS_3, "print_cross!3", &evaluator_rep::lib_print_cross_3);
KW (LIB_PRINT_TUPLE_1, MMX_PRINT_TUPLE, &evaluator_rep::lib_print_tuple_1);
KW (LIB_PRINT_TUPLE_2, "print_tuple!2", &evaluator_rep::lib_print_tuple_2);
KW (LIB_PRINT_TUPLE_3, "print_tuple!3", &evaluator_rep::lib_print_tuple_3);
KW (LIB_PRINT_ALIAS_1, MMX_PRINT_ALIAS, &evaluator_rep::lib_print_alias_1);
KW (LIB_PRINT_CODE_1, MMX_PRINT_CODE, &evaluator_rep::lib_print_code_1);

/******************************************************************************
* Extra routines
******************************************************************************/

#ifdef SUPPORTS_GC
generic
mem_inspect (generic g) {
  (void) g;
  gc_inspect ();
  return generic ();
}

generic
mem_memory_dump (generic g) {
  (void) g;
  gc_memory_dump ();
  return generic ();
}

generic
mem_show_useless (generic g) {
  (void) g;
  gc_show_objects (2);
  return generic ();
}

generic
mem_gc (generic g) {
  (void) g;
  interrupt_evaluation= true;
  return generic ();
}
#endif

generic
mem_used_bytes (generic g) {
  (void) g;
  return single (mmx_used_bytes ());
}

generic
mem_forget_sources (generic g) {
  (void) g;
  cur_location= 0;
  sources= list ();
  source_location= table ();
  return generic ();
}

/******************************************************************************
* Initialization
******************************************************************************/

void
evaluator_rep::declare (keyword kw) {
  env (kw->name)= kw;
}

evaluator_rep::evaluator_rep (): env (table ()) {
  cycle_witness= symbol ("begin!2");
#ifdef SUPPORTS_GC
  function inspect ("inspect", mem_inspect);
  declare ("inspect", value (inspect, function_type (ID_VOID)));
  function memory_dump ("memory_dump", mem_memory_dump);
  declare ("memory_dump", value (memory_dump, function_type (ID_VOID)));
  function show_useless ("show_useless", mem_show_useless);
  declare ("show_useless", value (show_useless, function_type (ID_VOID)));
  function gc ("gc", mem_gc);
  declare ("gc", value (gc, function_type (ID_VOID)));
#endif
  function used_bytes ("used_bytes", mem_used_bytes);
  declare ("used_bytes", value (used_bytes, function_type (ID_INT)));
  function forget_sources ("forget_sources", mem_forget_sources);
  declare ("forget_sources", value (forget_sources, function_type (ID_VOID)));

  declare (KW_COMPOUND_1);
  declare (KW_COMPOUND_2);
  declare (KW_COMPOUND_3);
  declare (KW_EVAL_ARGS_1);

  declare (KW_CONV_ALIASES_1);
  declare (KW_CONV_ALIASES_2);
  declare (KW_CONV_AMBIGUOUS_1);
  declare (KW_CONV_APPLY_1);
  declare (KW_CONV_APPLY_2);
  declare (KW_CONV_APPLY_3);
  declare (KW_CONV_CROSS_1);
  declare (KW_CONV_CROSS_2);
  declare (KW_CONV_FORALL_1);
  declare (KW_CONV_IGNORE_1);
  declare (KW_CONV_LAMBDA_1);
  declare (KW_CONV_SINGLET_1);
  declare (KW_CONV_SINGLET_2);
  declare (KW_CONV_TEMPLATE_1);
  declare (KW_CONV_TEMPLATE_2);
  declare (KW_CONV_TEMPLATE_3);
  declare (KW_CONV_UNALIAS_1);
  declare (KW_CONV_UNALIAS_2);
  declare (KW_CONV_UNAND_1);
  declare (KW_CONV_UNAND_2);
  declare (KW_CONV_UNAND_3);
  declare (KW_CONV_UNCROSS_1);
  declare (KW_CONV_UNCROSS_2);
  declare (KW_CONV_VCROSS_1);
  declare (KW_CONV_VCROSS_2);

  //declare (KW_ASSIGN_1);
  //declare (KW_ASSIGN_2);
  //declare (KW_ASSIGN_3);
  declare (KW_BEGIN_1);
  declare (KW_BEGIN_2);
  declare (KW_BREAK_1);
  declare (KW_CONTINUE_1);
  declare (KW_CROSS_1);
  declare (KW_CROSS_2);
  declare (KW_CASE_1);
  declare (KW_CASE_2);
  declare (KW_CASE_3);
  declare (KW_CASE_4);
  declare (KW_DEFINEALIAS_1);
  declare (KW_DEFINE_1);
  declare (KW_DEFINE_2);
  declare (KW_DEFINE_3);
  declare (KW_EXISTS_1);
  declare (KW_FOR_1);
  declare (KW_FORALL_1);
  declare (KW_FORALLVAL_1);
  declare (KW_FORALLVAL_2);
  declare (KW_FORIN_1);
  declare (KW_FORIN_2);
  declare (KW_FORIN_3);
  declare (KW_FORIN_4);
  declare (KW_FORIN_5);
  declare (KW_FORIN_6);
  declare (KW_FUNCTION_1);
  declare (KW_FUNCTION_2);
  declare (KW_GENERATE_1);
  declare (KW_GENERATE_2);
  declare (KW_GENERATE_3);
  declare (KW_IF_1);
  declare (KW_IF_2);
  declare (KW_JOIN_1);
  declare (KW_JOIN_2);
  declare (KW_JOIN_3);
  declare (KW_LAMBDA_1);
  declare (KW_LAMBDA_2);
  declare (KW_LOAD_1);
  declare (KW_LOAD_2);
  declare (KW_LOCAL_1);
  declare (KW_LOCAL_2);
  declare (KW_LOOP_1);
  declare (KW_LOOP_2);
  declare (KW_LOOP_3);
  declare (KW_LOOP_4);
  declare (KW_QUANTOR_2);
  declare (KW_RETRIEVE_1);
  declare (KW_RETRIEVE_2);
  declare (KW_RETRIEVE_3);
  declare (KW_RETURN_1);
  declare (KW_RETURN_2);
  declare (KW_SPLIT_1);
  declare (KW_SPLIT_2);
  declare (KW_STEP_1);
  declare (KW_SUPPORTS_1);
  declare (KW_SUPPORTS_2);
  declare (KW_TRANSTYPE_1);
  declare (KW_TRANSTYPE_2);
  declare (KW_TUPLE_1);
  declare (KW_TUPLE_2);
  declare (KW_UNTIL_1);
  declare (KW_USE_1);
  declare (KW_USE_2);
  declare (KW_VCROSS_1);
  declare (KW_VCROSS_2);
  declare (KW_VTUPLE_1);
  declare (KW_VTUPLE_2);
  declare (KW_WHILE_1);
  declare (KW_YIELD_1);
  declare (KW_YIELD_2);

  symbol T ("T");
  generic ID_OSTREAM_T_OSTREAM=
    function_type (ID_OSTREAM, T, ID_OSTREAM);
  generic ID_QT_OSTREAM_T_OSTREAM=
    template_type (T, ID_OSTREAM_T_OSTREAM);
  declare (MMX_LESSLESS_STR,
	   value (LIB_PRINT_ANY_1, ID_QT_OSTREAM_T_OSTREAM));
  mmx_lib_name (LIB_PRINT_ANY_1)= MMX_LESSLESS_STR;
}

#undef KW
} // namespace mmx
