
/******************************************************************************
* MODULE     : lib_print_any.cc
* DESCRIPTION: Printing any object (tuples and error handler)
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <evaluator.hh>
#include <printer.hh>
#include <match.hh>
namespace mmx {

void
evaluator_rep::lib_print_any_1 () {
  // Hack: pretend returning; necessary for library functions
  trace= cons (car (trace), cddr (trace));
  end_local_env ();
  // End hack

  CHECK_SYNC (MMX_PRINT_ANY);
  // Input: (print_any_1, arg, cont, ...)
  list    ret= cdr (as_list (car (rem)));
  printer out= as_printer (car (ret));
  generic cl = cadr (ret);
  generic tp = car (trace)[1][2];
  if (is_aliastype (tp)) {
    cur= LIB_PRINT_ALIAS_1;
    trace= cons (cons (MMX_PRINT_ALIAS, cdr (car (trace))), cdr (trace));
    // Output: (print_alias_1, arg, cont, ...)
    lib_print_alias_1 ();
  }
  else if (is_cross (tp) || is_vcross (tp)) {
    cur= LIB_PRINT_CROSS_1;
    rem= cons (out, cdr (cl), cdr (tp), cdr (rem));
    trace= cons (cons (MMX_PRINT_CROSS, cdr (car (trace))), cdr (trace));
    // Output: (print_cross_1, out, vals, types, cont, ...)
    lib_print_cross_1 ();
  }
  else if (is_tupletype (tp) || is_vtupletype (tp)) {
    cur= LIB_PRINT_TUPLE_1;
    rem= cons (out, cdr (cl), tp[1], cdr (rem));
    trace= cons (cons (MMX_PRINT_TUPLE, cdr (car (trace))), cdr (trace));
    // Output: (print_tuple_1, out, vals, type, cont, ...)
    lib_print_tuple_1 ();
  }
  else if (tp == ID_CLASS) {
    cur= LIB_PRINT_CLASS_1;
    trace= cons (cons (MMX_PRINT_CLASS, cdr (car (trace))), cdr (trace));
    // Output: (print_class_1, arg, cont, ...)
    lib_print_class_1 ();
  }
  else if (is_func (tp, ID_FORALL, 2) || is_func (tp, ID_FUNCTION, 2)) {
    cur= LIB_PRINT_CODE_1;
    trace= cons (cons (MMX_PRINT_CODE, cdr (car (trace))), cdr (trace));
    // Output: (print_code_1, arg, cont, ...)
    lib_print_code_1 ();
  }
  else {
    cerr << "  tp= " << tp << "\n";
    error (cadr (trace)[2], "No printing method for object");
  }
}

} // namespace mmx
