
/******************************************************************************
* MODULE     : extern.cc
* DESCRIPTION: for creating Mathemagix extensions
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <extern.hh>
#include <single.hh>
#include <function.hh>
#include <evaluator.hh>
namespace mmx {

GC_GLOBAL (evaluator, mmeval, evaluator ());

GC_GLOBAL_STATIC (table, initializers, table ());
GC_GLOBAL_STATIC (table, initialized, table ());

extern void initialize_base_types ();
extern void initialize_compound_types ();

/******************************************************************************
* Initializers
******************************************************************************/

void
register_initializer (string module, list files, function f) {
  while (!nil (files)) {
    generic key= list (module, car (files));
    initializers (key)= f;
    // cout << "  Registered " << module << ", " << car (files) << lf;
    files= cdr (files);
  }
}

void
call_initializer (string module, string file) {
  //cout << "Initializing " << module << ", " << file << lf << indent;
  generic key= list (module, file);
  if (!initializers->contains (key)) {
    //cout << unindent << "Not found" << lf;
    return;
  }
  function f= as_function (initializers [key]);
  if (initialized->contains (f)) {
    //cout << unindent << "Already done" << lf;
    return;
  }
  initialized (f)= single (true);
  //cout << "Calling " << f << " (" << file << ")" << lf;
  (void) f (generic ());
  //cout << unindent << "Initialization " << file << " completed" << lf;
}

void
initialize_glue () {
  list l= initializers -> entries ();
  while (!nil (l)) {
    string module= as_string (car (l) [0][0]);
    string file  = as_string (car (l) [0][1]);
    call_initializer (module, file);
    l= cdr (l);
  }
  initializers= table ();
  initialized = table ();
}

/******************************************************************************
* Type and function declarations
******************************************************************************/

void
declare (symbol var_name, generic val, generic type) {
  // cout << "    Declare " << var_name << ": " << type << " == " << val << lf;
  if (!nil (mmeval))
    mmeval->global_declare (var_name, value (val, type));
}

void
declare (symbol var_name, mmx_function fn, generic type) {
  // cout << "    Declare " << var_name << ": " << type << " == fun" << lf;
  function f (var_name, fn);
  if (!nil (mmeval))
    mmeval->global_declare (var_name, value (f, type));  
}

generic
template_type (generic v1, generic body) {
  return list (ID_FORALL, v1, body);
}

generic
template_type (generic v1, generic v2, generic body) {
  return template_type (v1, template_type (v2, body));
}

generic
function_type (generic r) {
  generic dom = list (ID_CROSS);
  return list (ID_FUNCTION, dom, r);
}

generic
function_type (generic x1, generic r) {
  generic dom = list (ID_CROSS, x1);
  return list (ID_FUNCTION, dom, r);
}

generic
function_type (generic x1, generic x2, generic r) {
  generic dom = list (ID_CROSS, x1, x2);
  return list (ID_FUNCTION, dom, r);
}

generic
function_type (generic x1, generic x2, generic x3, generic r) {
  generic dom = list (ID_CROSS, x1, x2, x3);
  return list (ID_FUNCTION, dom, r);
}

generic
tuple_function_type (generic x, generic r) {
  generic dom = list (ID_TUPLETYPE, x);
  return list (ID_FUNCTION, dom, r);
}

/******************************************************************************
* Main initialization
******************************************************************************/

void
initialize_master () {
  initialize_base_types ();
  initialize_compound_types ();
  initialize_glue ();
}

} // namespace mmx
