
/******************************************************************************
* MODULE     : ext_concrete.cc
* DESCRIPTION: Generate extension code for concrete exported types
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <extend.hh>
#include <ccpp.hh>
#include <pretty.hh>
namespace mmx {

/******************************************************************************
* Subroutines
******************************************************************************/

list
extender_rep::filter (list l, list target_l) {
  // Filter declarations and keep only those declarations which are exported
  if (nil (l)) return l;
  else {
    bool ok= false;
    generic var  = get_decl_var  (car (l));
    generic type = get_decl_type (car (l));
    if (export_env -> contains (var))
      ok= match_type (type, export_env [var]);
    if (ok) return cons (car (target_l), filter (cdr (l), cdr (target_l)));
    else return filter (cdr (l), cdr (target_l));
  }
}

generic
extender_rep::find_name (generic g) {
  // Find the name under which the routine is exported
  generic val= get_value (g, exports);
  if (nil (val)) {
    warning (g, "Assertion failed: could not find original export name");
    return g[2];
  }
  return val;
}

list
extender_rep::class_to_geninst (list l, int i) {
  // Substitute class arguments of templates by corresponding geninst_nr types
  if (nil (l)) return l;
  if (car (l) == MMX_CLASS)
    return cons (geninst (i), class_to_geninst (cdr (l), i+1));
  else return cons (car (l), class_to_geninst (cdr (l), i));
}

/******************************************************************************
* Produce the generic code for a given type
******************************************************************************/

list
extender_rep::contype_tmpl_args (list l, int kind, int i, int n) {
  if (nil (l)) return l;
  symbol  var1= string ("x_") * as_string (i);
  symbol  var2= string ("y_") * as_string (i);
  symbol  var3= string ("save_") * as_string (n);
  list code;
  generic type= car (l), tp;
  if (type == MMX_CLASS) tp= gentype ();
  else tp= type_as_ccpp (type);
  switch (kind) {
  case 1: // fields of type structure 
    code= Type (var1, tp);
    break;
  case 2: // arguments for constructor type structure
    code= Type (var2, tp);
    break;
  case 3: // initializers for fields of type structure
    code= generic (list (var1, var2));
    break;
  case 4: // for calling new operator
    code= generic (var1);
    break;
  case 5: // for marking routine
    if (type == MMX_CLASS)
      code= generic (list (symbol ("mark"), var1));
    break;
  case 6: // save old generic type arguments
    if (type == MMX_CLASS)
      code= Assign (Type (var3, gentype ()), curtype (n));
    break;
  case 7: // set generic type arguments
    if (type == MMX_CLASS)
      code= Assign (curtype (n), var1);
    break;
  case 8: // recover old generic type arguments
    if (type == MMX_CLASS)
      code= Assign (curtype (n), var3);
    break;
  }
  if (type == MMX_CLASS) n++;
  return code * contype_tmpl_args (cdr (l), kind, i+1, n);
}

generic
extender_rep::contype_cast (generic what, generic dtype, generic stype) {
  if (dtype == stype) return what;
  //return list (Specialize (symbol ("convert"), dtype, stype), what);
  if (dtype == geninst ()) return New (stype, what);
  if (stype == geninst ()) return Unpointer (Brackets (Transtype (Dot (what, symbol ("rep")), Pointer (dtype))));
  return Transtype (what, dtype);
}

list
extender_rep::contype_cast_args (generic ccpp_id, list l, int i) {
  if (nil (l)) return l;
  else {
    generic stype= car (l);
    generic dtype= substitute (stype, geninst (), ccpp_id);
    generic first= contype_cast (parameter (i), dtype, stype);
    return cons (first, contype_cast_args (ccpp_id, cdr (l), i+1));
  }
}

void
extender_rep::contype_routine (
  generic ccpp_id, generic ccpp_type, list args, generic g, generic name)
{
  //cout << "Generate method " << g << " for " << ccpp_type << lf;
  //cout << "  Exported as: " << name << lf;
  if (!(is_func (g, MMX_TYPE, 2) &&
	is_symbol (g[1]) &&
	is_func (g[2], MMX_FUNCTION, 2) &&
	is_func (g[2][1], MMX_CROSS)))
    error (g, "Assertion failed in contype_routine");
  else {
    // main computation
    generic rt= g[2][2];
    if (is_func (rt, MMX_CROSS, 0)) rt= MMX_VOID;
    list l1= cdr (as_list (g[2][1]));
    list l2= contype_cast_args (ccpp_type, l1, 1);
    generic op  = ccpp_back (name);
    if (op == symbol ("new")) op= ccpp_type_protect (ccpp_type);
    generic st  = substitute (rt, geninst (), ccpp_type);
    generic expr= contype_cast (cons (op, l2), rt, st);

    // special treatment of delete and assign operators
    if (op == symbol ("delete")) {
      assert (N (l1) == 1);
      generic arg= Dot (parameter (1), symbol ("rep"));
      generic ptr= Transtype (arg, Pointer (ccpp_type));
      expr= Delete (ptr);
    }
    if (op == symbol ("assign")) {
      generic ptr= Dot (parameter (1), symbol ("rep"));
      generic me = Transtype (ptr, Pointer (ccpp_type));
      expr= Assign (Unpointer (Brackets (me)), l2[1]);
    }

    // save generic types if necessary and treat void return values
    list code;
    list pre = contype_tmpl_args (args, 6) * contype_tmpl_args (args, 7);
    list post= contype_tmpl_args (args, 8);
    if (nil (pre) && nil (post)) {
      if (rt == MMX_VOID) code= list (expr);
      else code= list (Return (expr));
    }
    else {
      if (rt == MMX_VOID) code= pre * list (expr) * post;
      else {
	generic rvar= Type (symbol ("ret"), rt);
	generic comp= Assign (rvar, expr);
	code= pre * list (comp) * post * list (Return (symbol ("ret")));
      }
    }

    // generate declaration
    generic abbr_type_rep= contype_rep (ccpp_id);
    generic var = Scope (abbr_type_rep, g[1]);
    generic type= name_parameters (g[2]);
    generate (Define (Type (var, type), Begin (code)));
  }
}

void
extender_rep::contype_produce (
  generic mmx_id, generic ccpp_id, list args, int i,
  bool atomic_flag, bool top)
{
  // cout << "Produce " << mmx_id << ", " << ccpp_id << ", " << args << lf;
  generic abbr_type= contype (ccpp_id);
  generic abbr_type_rep= contype_rep (ccpp_id);
  generic mmx_type = mmx_id;
  generic ccpp_type= ccpp_id;

  // Pretreatment of template arguments
  list mmx_args = class_to_geninst (args);
  list ccpp_args= type_as_ccpp (mmx_args);
  if (!nil (args)) {
    mmx_type = cons (mmx_id , mmx_args);
    ccpp_type= cons (ccpp_id, ccpp_args);
  }
  list tl1= contype_tmpl_args (args, 1);
  list tl2= contype_tmpl_args (args, 2);
  list tl3= contype_tmpl_args (args, 3);
  list tl4= contype_tmpl_args (args, 4);
  list tl5= contype_tmpl_args (args, 5);

  // Produce the prototype
  generic ct= Define (Type (abbr_type_rep, Inline (Constructor (tl2))),
		      Initialize (cons (list (gentype_rep()), tl3), Begin()));
  generic in= Define (Type (symbol ("internal"),
			    Function (MMX_VOID, MMX_BOOL)),
		      Begin (Return (symbol ("true"))));
  list sp= specialize (gen_imports, mmx_type, false);
  list r= filter (sp, specialize (gen_imports, geninst (), false));
  list l= decl_as_ccpp (r, true);
  list x= list (ct, in);
  if (glue_flag) {
    generic t1= Function (MMX_VOID, symbol ("size_t"));
    generic e1= list (symbol ("sizeof"), abbr_type_rep);
    generic d1= Define (Type (symbol ("Size"), t1), Begin (Return (e1)));
    generic t2= Function (MMX_VOID, MMX_VOID);
    generic e2= list (Scope (gentype_rep (), symbol ("Mark")));
    generic d2= Define (Type (symbol ("Mark"), t2), Begin (cons (e2, tl5)));
    x= list (ct, d1, d2, in);
  }
  list mbs= tl1 * x * l;
  if (top) generate (make_struct (abbr_type_rep, mbs, list (gentype_rep ())));

  // Produce the constructor
  generic ft  = Function (Cross (tl1), gentype ());
  generic fun = Type (abbr_type, ft);
  generic ptr = New (abbr_type_rep, tl4);
  generic body= Begin (Return (ptr));
  if (glue_flag) {
    generic ret = symbol ("ret");
    list    mmxa= view_as_mmx (tl4, args, table ());
    generic back= mmx_make_list (cons (ext_id (i), mmxa));
    if (atomic_flag) back= ext_id (i);
    generic defp= Assign (Type (ret, Pointer (abbr_type_rep)), ptr);
    generic sett= Assign (Arrow (ret, symbol ("mmxt")), back);
    body= Begin (defp, sett, Return (ret));
  }
  if (top) generate (Define (fun, body));
  else generate (fun);

  // Produce the routines
  r= filter (sp, sp);
  while ((!nil (l)) && (!nil (r))) {
    if (top)
      contype_routine (ccpp_id, ccpp_type, args, car (l), find_name (car (r)));
    l= cdr (l);
    r= cdr (r);
  }
}

void
extender_rep::produce_contype (generic g, int i, bool top) {
  // cout << "Produce concrete type " << g << lf;
  if (!(is_func (g, MMX_DEFINE, 2) &&
	is_func (g[1], MMX_TYPE, 2) &&
	is_symbol (g[2])))
    error (g, "Assertion failed in produce_contype");
  else {
    generic type= g[1][2];
    if (type == MMX_CLASS)
      contype_produce (g[1][1], g[2], list (), i, true, top);
    else if (is_func (type, MMX_FUNCTION, 2) && (type[2] == MMX_CLASS)) {
      list l (type[1]);
      if (is_func (type[1], MMX_CROSS)) l= cdr (as_list (type[1]));
      contype_produce (g[1][1], g[2], l, i, false, top);
    }
  }
}

void
extender_rep::produce_contype (list l, int i) {
  if (!nil (l)) {
    produce_contype (car (l), i, ex_type_top->contains (car (l)));
    produce_contype (cdr (l), i+1);
  }
}

} // namespace mmx
