
/******************************************************************************
* MODULE     : ext_generic.cc
* DESCRIPTION: Produce the generic extension code
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <extend.hh>
#include <ccpp.hh>
#include <pretty.hh>
namespace mmx {

/******************************************************************************
* Specializing list of imports to a given type
******************************************************************************/

generic
extender_rep::specialize (generic g, generic to, bool virtual_flag) {
  // cout << "Specialize generic method " << g << " to " << to << lf;
  if (!(is_func (g, MMX_FORALL, 2) &&
	is_func (g[1], MMX_TYPE, 2) &&
	(g[1][2] == MMX_CLASS) &&
	is_func (g[2], MMX_DEFINE, 2) &&
	is_func (g[2][1], MMX_TYPE, 2)))
    error (g, "Assertion failed in specialize");
  else {
    generic qvar= g[1][1];
    generic body= g[2];
    generic ret = substitute (body, qvar, to);
    if (virtual_flag)
      ret= list (ret[0],
		 list (ret[1][0], ret[1][1], Virtual (ret[1][2])),
		 ret[2]);
    return ret;
  }
  return g;
}

list
extender_rep::specialize (list l, generic to, bool virtual_flag) {
  if (nil (l)) return l;
  else return cons (specialize (car (l), to, virtual_flag),
		    specialize (cdr (l), to, virtual_flag));
}

/******************************************************************************
* Produce the gentype type. This type serves as a base type for
* all dynamic types which have been exported by the user.
******************************************************************************/

static list
arg_list (int start, int end) {
  if (start >= end) return list ();
  else return cons (parameter (start), arg_list (start+1, end));
}

list
extender_rep::gentype_args_as_mmx (list l) {
  if (nil (l)) return l;
  generic var = car (l) [1];
  generic type= car (l) [2];
  generic rval= view_as_mmx (var, type, table ());
  generic rtype= build_mmx_type (type, list ());
  return cons (list (mmx_symbol ("value"), rval, rtype),
	       gentype_args_as_mmx (cdr (l)));
}

void
extender_rep::gentype_methods (generic h) {
  generic g= decl_as_ccpp (h);
  // cout << "Generate method " << g << " for " << gentype () << lf;
  if (!(is_func (g, MMX_TYPE, 2) &&
	is_symbol (g[1]) &&
	is_func (g[2], MMX_FUNCTION, 2) &&
	is_func (g[2][1], MMX_CROSS)))
    error (g, "Assertion failed in gentype_methods");
  else {
    g= name_parameters (g);
    generic name= genmethod (g[1]);
    generic scoped= Scope (gentype_rep (), name);
    generic fun= list (g[0], scoped, g[2]);
    if (glue_flag) {
      h= name_parameters (h[1]);
      generic back= build_mmx_type (h[2][2], list ());
      back= substitute (back, geninst (), symbol ("mmxt"));
      back= list (mmx_symbol ("value"), back, mmx_symbol ("ID_CLASS"));
      string sname= as_string (S (as_symbol (h[1])));
      generic symb= list (mmx_symbol ("symbol"), sname);
      if (sname == "new") symb= list (mmx_symbol ("MMX_NEW"), back[1]);
      list args= gentype_args_as_mmx (cdr (as_list (h[2][1])));
      generic instr= mmx_make_list (cons (symb, args));
      instr= list (mmx_symbol ("list"), mmx_symbol ("MMX_TRANSTYPE"),
		   instr, back);
      generic body= list (mmx_symbol ("extern_eval"), instr);
      if (g[2][2] != MMX_VOID) {
	generic ret= view_as_ccpp (body, h[2][2], table ());
	body= Return (ret);
      }
      if (sname == "delete") {
	generic ptr= Dot (parameter (1), symbol ("rep"));
	body= Delete (Transtype (ptr, Pointer (mmx_symbol ("external"))));
      }
      if (sname == "assign") {
	generic ext= mmx_symbol ("external");
	generic ptr= Dot (parameter (1), symbol ("rep"));
	generic me = Transtype (ptr, Pointer (ext));
	body= Assign (Unpointer (Brackets (me)), list (ext, args[1][1]));
      }
      generate (Define (fun, Begin (body)));
    }
    else {
      string err ("\"Fatal error: Generic method " *
		  as_string (S (as_symbol (g[1]))) *
		  " is not supported\\n\"");
      generic _error= list (MMX_LESSLESS, symbol ("cerr"), symbol (err));
      generic _exit = list (symbol ("exit"), symbol ("1"));
      generate (Define (fun, Begin (_error, _exit)));
    }
  }
}

void
extender_rep::gentype_methods (list l) {
  if (!nil (l)) {
    gentype_methods (car (l));
    gentype_methods (cdr (l));
  }
}

void
extender_rep::produce_gentype () {
  // Default methods of the gentype_rep structure
  list l= specialize (gen_imports, geninst (), true);
  l= decl_as_ccpp (l, true);
  generic gt= Type (symbol ("mmxt"), mmx_generic ());
  generic md= Type (symbol ("g"), mmx_generic ());
  generic mi= list (symbol ("mmxt"), symbol ("g"));
  generic mc= Define (Type (gentype_rep (), Inline (Constructor (md))),
		      Initialize (mi, Begin ()));
  generic gc= Define (Type (gentype_rep (), Inline (Constructor (MMX_VOID))),
		      Begin ());
  generic gd= Define (Type (gentype_rep (), Inline (Virtual (Destructor ()))),
		      Begin ());
  generic in= Define (Type (symbol ("internal"),
			    Virtual (Function (MMX_VOID, MMX_BOOL))),
		      Begin (Return (symbol ("false"))));

  if (glue_flag) {
    // Implement virtual methods inherited from generic_rep
    generic g = symbol ("g");
    generic o = symbol ("out");
    generic t = symbol ("mmxt");
    generic r = symbol ("rep");
    generic t1= Virtual (Function (MMX_VOID, symbol ("size_t")));
    generic e1= list (symbol ("sizeof"), gentype_rep ());
    generic d1= Define (Type (symbol ("Size"), t1), Begin (Return (e1)));
    generic t2= Virtual (Function (MMX_VOID, MMX_VOID));
    generic e2= list (symbol ("mark"), t);
    generic d2= Define (Type (symbol ("Mark"), t2), Begin (e2));
    generic t3= Function (MMX_VOID, MMX_INT);
    generic e3= symbol ("GLUED_TYPE_TYPE");
    generic d3= Define (Type (symbol ("Type"), t3), Begin (Return (e3)));
    generic t4= Function (MMX_VOID, MMX_INT);
    generic e4= list (symbol ("hash"), t);
    generic d4= Define (Type (symbol ("Hash"), t4), Begin (Return (e4)));
    generic t5= Function (mmx_generic (), MMX_BOOL);
    generic d5= Type (symbol ("Equal"), t5);
    generic t6= Function (Type (o, Alias (MMX_OSTREAM)), MMX_VOID);
    generic e6= list (symbol ("print"), o, string ("#glued_type"), t);
    generic d6= Define (Type (symbol ("Print"), t6), Begin (e6));
    gd= Define (Type (gentype_rep (), Inline (Destructor ())), Begin ());
    l= cons (gt, gc, mc, gd,
	     cons (d1, d2, d3, d4,
		   cons (d5, d6, in, l)));
    list inh (Public (mmx_symbol ("generic_rep")));
    generate (make_struct (gentype_rep (), l, inh));

    // Generate the is_gentype predicate
    generic pt= Inline (Function (Type (g, mmx_generic ()), MMX_BOOL));
    generic tt= list (symbol ("gentype"), g);
    generic pr= list (MMX_EQUAL, tt, symbol ("GLUED_TYPE_TYPE"));
    generate (Define (Type (symbol ("is_") * gentype (), pt),
		      Begin (Return (pr))));

    // Generate the gentype structure
    generic s1= list (symbol ("DEFAULT_METHODS"), gentype ());
    generic i2= list (r, New (gentype_rep ()));
    generic s2= Define (Type (gentype (), Inline (Constructor ())),
			Initialize (i2, Begin ()));
    generic i3= list (r, New (gentype_rep (), g));
    generic s3= Define (Type (gentype (), Inline (Constructor (md))),
			Initialize (i3, Begin ()));
    generic st= make_struct (gentype (), list (s1, s2, s3));
    generic gl= list (symbol ("TMPL_GC_REGISTER_TYPE"), gentype ());
    generate (Namespace (symbol ("mmx"), Begin (st, gl)));
    generate (Using (mmx_scope (gentype ())));

    // The Equal method should be specified here since it refers to gentype
    if (gen_import_top) {
      generic T5= Function (Type (g, mmx_generic ()), MMX_BOOL);
      generic c1= list (symbol ("is_glued_type"), g);
      generic as= symbol ("as_") * gentype ();
      generic c2= list (MMX_EQUAL, t, Arrow (list (as, g), t));
      generate (Define (Type (Scope (gentype_rep(), symbol ("Equal")), T5),
			Begin (Return (list (MMX_SEQAND, c1, c2)))));
    }

    // Conversion between gentype and Mathemagix types
    if (!nil (gen_imports)) {
      if (gen_import_top) {
	generic tcv= Specialize (mmx_symbol ("type_converter"), gentype ());
	generate (Global (Type (ext_cv (), tcv)));
      }
      else {
	generic tcv= list (mmx_symbol ("type_converter"), gentype ());
	generate (Type (ext_cv (), Extern (tcv)));
      }
    }
    generate (list (symbol ("GENTYPE_CONVERTERS"),
		    ext_cv (), gentype (), gentype_rep ()));
  }
  else {
    l= cons (gc, gd, in, l);
    generate (make_struct (gentype_rep (), l));

    // The gentype is a pointer to gentype_rep
    generate (Type (gentype_rep (), MMX_CLASS));
    generate (Typedef (gentype (), Pointer (gentype_rep ())));
  }

  // Printing geninst instances should never happen,
  // but the routine should be included to make the compiler happy
  if (glue_flag) {
    generic x  = symbol ("g");
    generic out= symbol ("out");
    generic ost= Alias (MMX_OSTREAM);
    generic dom= Cross (Type (out, ost), Type (x, geninst ()));
    generic fun= Type (MMX_LESSLESS, Function (dom, ost));
    string err ("\"Fatal error: << should not be called for generics\"");
    generic _error= list (MMX_LESSLESS, symbol ("cerr"), symbol (err));
    generic _exit = list (symbol ("exit"), symbol ("1"));
    if (gen_import_top) generate (Define (fun, Begin (_error, _exit)));
    else generate (fun);
  }

  // Generate the methods
  l= specialize (gen_imports, geninst (), false);
  if (gen_import_top) gentype_methods (l);
}

/******************************************************************************
* Produce the unnumbered generic instance types
******************************************************************************/

/*
static generic
converter_class (
  generic tmpl_var, generic cvr, generic F, generic T, generic body)
{
  generic x  = symbol ("x");
  generic h  = symbol ("helper");

  // the contents of the converter structure
  generic tx = Type (x, F);
  generic f  = Static (Inline (Function (tx, T)));
  generic b  = Define (Type (h, f), body);
  generic val= Struct (b);

  // the converter class with helper routine
  generic var= Specialize (cvr, T, F);
  generic tp = Forall (Type (tmpl_var, MMX_CLASS), MMX_CLASS);
  if (N(tmpl_var) == 2) {
    if ((F == symbol ("F")) && (T == symbol ("T"))) var= cvr;
    tp = Forall (Type (tmpl_var[0], MMX_CLASS),
		 Forall (Type (tmpl_var[1], MMX_CLASS),
			 MMX_CLASS));
  }
  return Define (Type (var, tp), val);
}
*/

void
extender_rep::produce_geninst () {
  generate (make_blackbox_struct (geninst ()));
  //if (glue_flag)
  //generate (list (symbol ("GENINST_CONVERTERS"), geninst ()));

  //generic x  = symbol ("x");
  //generic h  = symbol ("helper");
  //generic cv = symbol ("convert");
  //generic cvr= symbol ("converter");
  //generic F  = symbol ("F");
  //generic T  = symbol ("T");

  /*
  // The converter meta template
  generic b0= Begin (Return (Transtype (x, T)));
  generate (converter_class (list (T, F), cvr, F, T, b0));

  // From abstract to concrete
  generic a1= Transtype (Dot (x, symbol ("rep")), Pointer (T));
  generic b1= Begin (Return (Unpointer (Brackets (a1))));
  generate (converter_class (T, cvr, geninst (), T, b1));

  // From concrete to abstract
  generic b2= Begin (Return (list (geninst (), New (F, x))));
  generate (converter_class (F, cvr, F, geninst (), b2));

  // The main conversion routine
  generic cc= Scope (list (cvr, T, F), h);
  generic ff= Inline (Function (Type (x, F), T));
  generic tt= Forall (Type (T, MMX_CLASS), Forall (Type (F, MMX_CLASS), ff));
  generic bb= Return (list (cc, x));
  generate (Define (Type (cv, tt), Begin (bb)));
  */
}

/******************************************************************************
* Produce the generic instance types, indexed by numbers 1, 2, ....
* These types are instances of dynamic types. curinst(i) holds
* the current type of geninst(i).
******************************************************************************/

static generic
Constify (generic g, generic type) {
  assert (is_func (g, MMX_FUNCTION, 2));
  generic dom= substitute (g[1], type, Const (Alias (type)));
  return list (g[0], dom, g[2]);
}

list
extender_rep::abstract (int nr, list args, list types) {
  // Convert to abstract type when necessary
  if (nil (args)) return args;
  generic x= car (args);
  generic stype= car (types) [2];
  generic dtype= substitute (stype, geninst (nr), geninst ());
  if (dtype != stype)
    x= list (Specialize (symbol ("cast"), dtype, stype), x);
  return cons (x, abstract (nr, cdr (args), cdr (types)));
}

void
extender_rep::geninst_constructors (int nr, generic g, list& r) {
  //cout << "Generate constructor " << g << " for " << geninst (nr) << lf;
  if (!(is_func (g, MMX_TYPE, 2) &&
	is_func (g[2], MMX_FUNCTION, 2) &&
	is_func (g[2][1], MMX_CROSS)))
    {
      error (g, "Assertion failed in geninst_constructors");
      return;
    }

  bool copy_flag= (g[2][1] == Cross (g[2][2]));
  if (g[1] == symbol ("assign"))
    g= list (g[0], g[1], Function (Cross (cddr (as_list (g[2][1]))), g[2][2]));
  g= name_parameters (g);
  generic rep = symbol ("rep");
  generic name= genmethod (g[1]);
  generic fun = Arrow (curtype (nr), name);
  if (g[1] == symbol ("new")) {
    list    args= arg_list (1, N(g[2][1]));
    generic appl= cons (fun, abstract (nr, args, cdr (as_list (g[2][1]))));
    if (copy_flag)
      appl= list (fun, list (geninst(), Dot (parameter (1), rep)));
    generic inst= Dot (appl, rep);
    generic body= Initialize (generic (list (rep, inst)), Begin ());
    generic tps = cdr (as_list (g[2][1]));
    generic sts = substitute (tps, geninst (nr), Const (Alias (geninst (nr))));
    generic type= Inline (Constructor (as_list (sts)));
    r= cons (Define (list (g[0], geninst (nr), type), body), r);
  }
  else if (g[1] == symbol ("delete")) {
    generic arg = list (geninst (), rep);
    generic appl= list (fun, arg);
    generic body= Begin (appl);
    generic type= Inline (Destructor ());
    r= cons (Define (list (g[0], geninst (nr), type), body), r);
  }
  else if (g[1] == symbol ("assign")) {
    list    args= arg_list (1, N(g[2][1]));
    list    bis = abstract (nr, args, cdr (as_list (g[2][1])));
    generic appl= cons (fun, list (geninst (), rep), bis);
    generic body= Begin (appl, Return (Unpointer (symbol ("this"))));
    list    tps = cdr (as_list (g[2][1]));
    generic fn  = Function (Cross (tps), Alias (geninst (nr)));
    generic type= Inline (Constify (fn, geninst (nr)));
    r= cons (Define (list (g[0], symbol ("operator ="), type), body), r);
  }
}

list
extender_rep::geninst_constructors (int nr, list l) {
  list r;
  while (!nil (l)) {
    geninst_constructors (nr, car (l), r);
    l= cdr (l);
  }
  return reverse (r);
}

void
extender_rep::geninst_methods (int nr, generic g) {
  // cout << "Generate method " << g << " for " << geninst (nr) << lf;
  if (!(is_func (g, MMX_TYPE, 2) &&
	is_symbol (g[1]) &&
	is_func (g[2], MMX_FUNCTION, 2) &&
	is_func (g[2][1], MMX_CROSS)))
    error (g, "Assertion failed in geninst_methods");
  else if (g[1] == symbol ("new"));
  else if (g[1] == symbol ("delete"));
  else if (g[1] == symbol ("assign"));
  else {
    g= name_parameters (g);
    generic name = genmethod (g[1]);
    generic fun  = Arrow (curtype (nr), name);
    list    args = arg_list (1, N(g[2][1]));
    generic appl = cons (fun, abstract (nr, args, cdr (as_list (g[2][1]))));
    generic dtype= g[2][2];
    generic stype= substitute (dtype, geninst (nr), geninst ());
    generic conv = appl;
    if (dtype != stype)
      conv= list (Specialize (symbol ("cast"), dtype, stype), conv);
    generic prot= list (g[0], g[1], Inline (Constify (g[2], geninst (nr))));
    if (dtype == MMX_VOID) generate (Define (prot, Begin (conv)));
    else if (dtype == geninst (nr)) {
      // NOTE: the following hack is used as the matching "delete" for
      // the "new" geninst object created by the corresponding genmethod.
      // We copy the result in a geninst_nr object and delete the original.
      // The geninst_nr object is deleted by C++ (really?) only after use.
      generic aux = symbol ("aux");
      generic auxd= Assign (Type (aux, geninst ()), appl);
      generic ret = symbol ("ret");
      generic conv= list (Specialize (symbol ("cast"), dtype, stype), aux);
      generic retd= Assign (Type (ret, geninst (nr)), conv);
      generic del = list (Arrow (curtype (nr), symbol ("gen_delete")), aux);
      generate (Define (prot, Begin (auxd, retd, del, Return (ret))));
    }
    else generate (Define (prot, Begin (Return (conv))));
  }
}

void
extender_rep::geninst_methods (int nr, list l) {
  if (!nil (l)) {
    geninst_methods (nr, car (l));
    geninst_methods (nr, cdr (l));
  }
}

void
extender_rep::produce_geninst (int nr) {
  list l = specialize (gen_imports, geninst (nr), false);
  l      = decl_as_ccpp (l);
  list cl= geninst_constructors (nr, l);
  generate (make_blackbox_struct (geninst (nr), cl));
  if (glue_flag)
    generate (list (symbol ("GENINSTX_CONVERTERS"),
		    geninst (nr), curtype (nr)));
  geninst_methods (nr, l);
}

} // namespace mmx
