
/******************************************************************************
* MODULE     : extend.hh
* DESCRIPTION: Produce dynamic glue libraries for extern C++ template libraries
* COPYRIGHT  : (C) 2004  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __EXTEND_HH
#define __EXTEND_HH
#include <symbols.hh>
#include <list.hh>
#include <table.hh>
namespace mmx {

#define MAX_TYPE_ARGS 9

class extender_rep: generic_rep {
private:
  string the_prefix;        // prefix for the abstract signature space
  string cur_file;          // current file name
  bool   top_level;         // are we parsing a top level file or an inclusion?
  table  loaded;            // all files which have currently been loaded
  list   top_files;         // list of top files for auto initializer
  bool   glue_flag;         // produce glue code for mathematics
  int    nr_errors;         // number of errors
  string preamble;          // C++ preamble
  string postamble;         // C++ postamble
  list   gen_imports;       // list with routines for the generic type
  bool   gen_import_start;  // did we start to define generic routines?
  bool   gen_import_top;    // are the generic routines top-level?
  table  gen_import_env;    // contains the type of each generic routine
  list   ex_types;          // list with exported type declarations
  table  ex_type_top;       // is an exported type declaration top-level?
  table  ex_type_conv;      // associates Mmx name to C++ name
  list   exports;           // list with all exported routines
  table  export_top;        // is an exported routine top-level?
  table  export_env;        // contains the type of each exported routine
  list   im_types;          // [unused] list with imported type declarations
  table  im_type_top;       // [unused] is imported type declaration top-level?
  table  im_type_conv;      // [unused] associates Mmx name to C++ name
  list   imports;           // [unused] all imported routines
  table  import_top;        // [unused] is an imported routine top-level?
  table  import_env;        // [unused] the type of each imported routine
  list   output;            // the generated output
  list   glue_init;         // the initialization routine for the glue
  int    max_type_args;     // maximal number of type arguments of templates
  int    nr_ext_tvar;       // maximal number of Mmx template arguments

protected:
  size_t Size ();
  void Mark ();
  int Type ();
  void Print (ostream& out);

  // remove naming conflict and remake global routines visible
  inline generic Access (generic g1, generic g2) {
    return list (MMX_ACCESS, g1, g2); }
  inline generic Type (generic g1, generic g2) {
    return list (MMX_TYPE, g1, g2); }

private:
  // further names
  symbol  gentype ();
  symbol  gentype_rep ();
  generic genmethod (generic name);
  generic geninst ();
  generic geninst (int nr);
  generic curtype (int nr);
  generic contype (generic name);
  generic contype_rep (generic name);

  generic mmx_scope (generic g);
  generic mmx_symbol (symbol s);
  generic mmx_specialize (symbol s, generic a1);
  generic mmx_specialize (symbol s, generic a1, generic a2);
  generic mmx_specialize (symbol s, generic a1, generic a2, generic a3);
  generic mmx_generic ();
  generic mmx_mirror ();
  symbol  ext_cv ();
  symbol  ext_id (int i);
  symbol  ext_fun (int i);
  symbol  ext_conv (int i);
  symbol  ext_tvar (int i);
  symbol  ext_arg (int i);

  // error handling, code generation, and other helper routines
  void    error (generic where, string message);
  void    warning (generic where, string message);
  void    generate (generic g);
  void    generate_init (generic g);
  void    declare (generic var, generic val, generic type);
  generic mmx_make_list (list l);

  // parsing
  generic rewrite (generic g);
  void    ex_type_decl (generic g);
  void    gen_import_decl (generic g);
  void    export_decl (generic g);
  list    add_default_routines (generic type, mmx_function fn);
  void    add_default_routines ();
  generic load (string name);
  void    parse (generic prog, bool reverse_flag= true);
  void    print_dependencies (generic prog);

  // conversions
  generic type_as_ccpp (generic g, table t= table ());
  list    type_as_ccpp (list l, table t= table ());
  generic make_mmx_gentype (generic g);
  generic type_as_expr (generic g, bool flag);
  generic decl_as_ccpp (generic g, bool gen_flag= false);
  list    decl_as_ccpp (list l, bool gen_flag= false);
  generic view_as_ccpp (generic expr, generic mmx_type, table t= table ());
  generic view_as_mmx (generic expr, generic mmx_type, table t);
  list    view_as_mmx (list exprs, list mmx_types, table t);

  // extension code for generic types
  generic specialize (generic g, generic to, bool virtual_flag);
  list    specialize (list l, generic to, bool virtual_flag);
  list    gentype_args_as_mmx (list l);
  void    gentype_methods (generic g);
  void    gentype_methods (list l);
  void    produce_gentype ();
  void    produce_geninst ();
  list    abstract (int nr, list args, list types);
  void    geninst_constructors (int nr, generic g, list& r);
  list    geninst_constructors (int nr, list l);
  void    geninst_methods (int nr, generic g);
  void    geninst_methods (int nr, list l);
  void    produce_geninst (int nr);

  // extension code for concrete types
  list    filter (list l, list target_l);
  generic find_name (generic g);
  list    class_to_geninst (list l, int i=1);
  list    contype_tmpl_args (list l, int kind, int i=1, int n=1);
  generic contype_cast (generic what, generic dtype, generic stype);
  list    contype_cast_args (generic tp, list l, int i);
  void    contype_routine (generic n, generic t, list a, generic g, generic h);
  void    contype_produce (generic mid, generic cid, list ta, int i,
			   bool f, bool top);
  void    produce_contype (generic g, int i, bool top);
  void    produce_contype (list l, int i);

  // generation of the glue
  bool    match_mmx_type (generic expr, generic defn, list bindings);
  int     resolve_mmx_type (generic expr, list l, int i, list bindings);
  generic build_mmx_type_sub (generic g, list bindings);
  list    build_mmx_type (list l, list bindings);
  generic build_mmx_type (generic g, list bindings);
  void    produce_glue_type (generic g, int i, bool prototype);
  void    produce_glue_type (list l, int i, bool prototype);
  list    glue_template_args (list l, int kind, int i=1, int n=1);
  list    glue_convert_args (list l, generic x, table t, int i);
  void    produce_glue_routine (generic g, int i, list bindings);
  void    produce_glue_routine (list l, int i);
  void    produce_glue ();

public:
  inline extender_rep (string prefix):
    the_prefix (prefix),
    cur_file (""),
    top_level (true),
    glue_flag (true),
    nr_errors (0),
    gen_import_start (false),
    max_type_args (0),
    nr_ext_tvar (0) {}
  inline ~extender_rep () {}

  void produce (ostream& out, generic prog);

  friend class extender;
  friend void print_dependencies (string ofile, string prefix, generic prog);
  friend bool extension_code (ostream& out, string pre, generic prog, bool f);
};

class extender {
DEFAULT_METHODS(extender);
public:
  inline extender (string prefix): rep (new extender_rep (prefix)) {}
};

inline symbol operator * (symbol s1, symbol s2) {
  return symbol (as_string (S (s1)) * as_string (S (s2))); }
void nice_print (generic text, list l);
void nice_print (symbol text, list l);

generic get_decl_var (generic g);
generic get_decl_type (generic g);
generic get_decl_val (generic g);
bool    match_type (generic what, generic in, table t= table ());
generic get_value (generic what, list l);
generic make_struct (generic name, list args, list inherit= list ());
generic make_blackbox_struct (generic name, list extra= list ());
generic ccpp_type_protect (generic type);
symbol  parameter (int i);
generic name_parameters (generic name);
generic Global (generic decl);

}
#endif // __EXTEND_HH
