
/******************************************************************************
* MODULE     : std_error.cc
* DESCRIPTION: Standard error messages
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <error.hh>
namespace mmx {

class std_error_rep: public generic_rep {
protected:
  string  message;
  generic src;

protected:
  size_t Size ();
  void Mark ();
  int Type () { return ERROR_TYPE; }
  void Print (ostream& out);

public:
  inline std_error_rep (string message2, generic src2) :
    message (message2), src (src2) { index (src2); }
};

size_t
std_error_rep::Size () {
  return sizeof (std_error_rep);
}

void
std_error_rep::Mark () {
  mark (message);
  mark (src);
}

void
std_error_rep::Print (ostream& out) {
  if (current_print_format == normal) {
    print_error_location (out, src);
    out << message << lf;
    print_error_source (out, src);
  }
  else out << "#std_error<" << message << ">";
}

generic
std_error (string message, generic src) {
  return new std_error_rep (message, src);
}

} // namespace mmx
