
/******************************************************************************
* MODULE     : converter.hh
* DESCRIPTION: converters of type expressions
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __CONVERTER_HH
#define __CONVERTER_HH
#include <match.hh>
#include <proof.hh>
#include <array.hh>
namespace mmx {

#define CV_NORMAL   0
#define CV_DIRECT   1
#define CV_COLLECT  2

class converter_rep: public generic_rep {
public:
  generic dcv;     // direct implicit converters
  generic lcv;     // left associative implicit converters
  generic rcv;     // right associative implicit converters

protected:
  size_t Size ();
  void Mark ();
  int Type ();
  int Hash ();
  bool Equal (generic g);
  generic Copy ();
  int Arity ();
  void Print (ostream& out);

public:
  converter_rep ();
  void add (generic what, bool left_ass, bool right_ass);

  bool freely_converts (generic from, generic to, int mode);
  list disambiguate (list deps, int mode);
  list disambiguate_hack (list deps, int mode);

  proof convert_from_and (generic from, generic to, int mode);
  proof convert_from_or (generic from, generic to, int mode);
  proof convert_from_forall (generic from, generic to, int mode);
  proof convert_from_exists (generic from, generic to, int mode);
  proof convert_to_and (generic from, generic to, int mode);
  proof convert_to_or (generic from, generic to, int mode);
  proof convert_to_forall (generic from, generic to, int mode);
  proof convert_to_exists (generic from, generic to, int mode);

  proof convert_function (generic from, generic to, int mode);
  proof convert_cross (generic from, generic to, int mode);
  proof convert_tuple (generic from, generic to, int mode);
  proof convert_ignore (generic from, generic to, int mode);
  proof convert_from_singlet (generic from, generic to, int mode);
  proof convert_to_singlet (generic from, generic to, int mode);
  proof convert_from_alias (generic from, generic to, int mode);
  proof convert_aliases (generic from, generic to, int mode);

  proof apply (generic f, proof to, int mode);
  proof pullback (generic f, proof to, int mode);
  proof convert_implicit (generic from, generic to, int mode);

  proof convert (generic from, generic to, int mode);

  generic apply_prog (generic f, generic arg);
  friend generic convert_prepare (generic g);
  void test (generic cmd);
  friend class converter;
};

class converter {
DEFAULT_METHODS(converter);
public:
  inline converter (): rep (new converter_rep ()) {}

  inline bool operator == (converter cv) { return (rep == cv.rep); }
  inline bool operator != (converter cv) { return (rep != cv.rep); }

  friend converter copy (converter cv);
  friend ostream& operator << (ostream& out, converter cv);
};

}
#endif // __CONVERTER_HH
