
/******************************************************************************
* MODULE     : quantor.hh
* DESCRIPTION: quantified variables, identified by unique numbers
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __QUANTOR_HH
#define __QUANTOR_HH
#include <symbol.hh>
#include <table.hh>
#include <list.hh>
namespace mmx {

extern long long int qt_serial;

enum quantor_type {
  qt_free,    // a free variable obtained from a universal quantifier
  qt_witness, // a witness for an existential quantifier
  qt_bound    // a bound variable after requantification
};

NO_MARKING(quantor_type);

class quantor_rep: public generic_rep {
public:
  quantor_type  type;  // type of variable
  long long int nr;    // serial number
  generic       name;  // its name

protected:
  size_t Size ();
  void Mark ();
  int Type ();
  int Hash ();
  bool Equal (generic g);
  generic Copy ();
  int Arity ();
  void Print (ostream& out);

public:
  inline quantor_rep (quantor_type type2, long long int nr2, generic name2):
    type (type2), nr (nr2), name (name2) {}
  friend class quantor;
};

class quantor {
DEFAULT_METHODS(quantor);
public:
  inline quantor (quantor_type type, long long int nr, generic name):
    rep (new quantor_rep (type, nr, name)) {}

  inline bool operator == (quantor v) {
    return (rep->type == v->type) && (rep->nr == v->nr); }
  inline bool operator != (quantor v) {
    return (rep->type != v->type) || (rep->nr != v->nr); }

  friend quantor copy (quantor v);
  friend ostream& operator << (ostream& out, quantor v);
};

inline quantor free_var (generic name) {
  return quantor (qt_free, qt_serial++, name); }
inline quantor witness_var (generic name, long long int nr) {
  return quantor (qt_witness, nr, name); }
inline quantor witness_var (generic name) {
  return quantor (qt_witness, qt_serial++, name); }
inline quantor bound_var (generic name) {
  return quantor (qt_bound, qt_serial++, name); }
inline bool is_free_var (generic g) {
  return is_quantor (g) && (as_quantor (g)->type == qt_free); }
inline bool is_witness_var (generic g) {
  return is_quantor (g) && (as_quantor (g)->type == qt_witness); }
inline bool is_bound_var (generic g) {
  return is_quantor (g) && (as_quantor (g)->type == qt_bound); }

generic highest_in (generic f);

}
#endif // __QUANTOR_HH
