
/******************************************************************************
* MODULE     : ccpp.cc
* DESCRIPTION: Convert mathemagix trees to C or C++
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <symbols.hh>
#include <list.hh>
#include <table.hh>
#include <serializer.hh>
namespace mmx {

GC_GLOBAL (symbol, CCPP_ARROW, "->%");
GC_GLOBAL (symbol, CCPP_CAST, "cast%");
GC_GLOBAL (symbol, CCPP_CONSTRUCT, "construct%");
GC_GLOBAL (symbol, CCPP_CONSTRUCTOR, "constructor%");
GC_GLOBAL (symbol, CCPP_DELETE, "delete%");
GC_GLOBAL (symbol, CCPP_DESTRUCTOR, "destructor%");
GC_GLOBAL (symbol, CCPP_INHERIT, "inherit%");
GC_GLOBAL (symbol, CCPP_INITIALIZE, ":%");
GC_GLOBAL (symbol, CCPP_NAMESPACE, "namespace%");
GC_GLOBAL (symbol, CCPP_NEW, "new%");
GC_GLOBAL (symbol, CCPP_SCOPE, "::%");
GC_GLOBAL (symbol, CCPP_SPECIALIZE, ".<>%");
GC_GLOBAL (symbol, CCPP_STRUCT, "struct%");
GC_GLOBAL (symbol, CCPP_TYPEDEF, "typedef%");
GC_GLOBAL (symbol, CCPP_UNALIAS, "unalias%");
GC_GLOBAL (symbol, CCPP_UNPOINTER, "unpointer%");
GC_GLOBAL (symbol, CCPP_UNFUNCTION, "unfunction%");
GC_GLOBAL (symbol, CCPP_USING, "using%");
GC_GLOBAL (symbol, CCPP_VERBATIM, "verbatim%");

GC_GLOBAL (symbol, CCPP_CONST, "Const");
GC_GLOBAL (symbol, CCPP_EXTERN, "Extern");
GC_GLOBAL (symbol, CCPP_INLINE, "Inline");
GC_GLOBAL (symbol, CCPP_PUBLIC, "Public");
GC_GLOBAL (symbol, CCPP_STATIC, "Static");
GC_GLOBAL (symbol, CCPP_VIRTUAL, "Virtual");

GC_GLOBAL (symbol, CCPP_BRACKETS, "brackets%");

GC_GLOBAL_STATIC (table, op_back, table ());
GC_GLOBAL_STATIC (table, op_name, table ());
GC_GLOBAL_STATIC (table, op_prot, table ());

/******************************************************************************
* Useful symbols and subroutines
******************************************************************************/

static void
add (symbol var, symbol val, symbol prot) {
  op_back (var)= var;
  op_name (var)= val;
  op_prot (var)= prot;
  op_back (val)= var;
  op_name (val)= val;
  op_prot (val)= prot;
  op_back (symbol (string ("operator ") * as_string (S (val))))= var;
  op_name (symbol (string ("operator ") * as_string (S (val))))= val;
  op_prot (symbol (string ("operator ") * as_string (S (val))))= prot;
}

static void
ccpp_init () {
  if (N(op_name) == 0) {
    add (MMX_PLUS, "+", "plus");
    add (MMX_MINUS, "-", "minus");
    add (MMX_TIMES, "*", "times");
    add (MMX_OVER, "/", "over");
    add (MMX_MOD, "%", "mod");
    add (MMX_XOR, "^", "xor");
    add (MMX_AND, "&", "and");
    add (MMX_OR, "|", "or");
    add (MMX_NOT, "!", "not");
    add (MMX_ASSIGN, "=", "assign");
    add (MMX_LESS, "<", "less");
    add (MMX_GTR, ">", "gtr");
    add (MMX_PLUSASSIGN, "+=", "plusassign");
    add (MMX_MINUSASSIGN, "-=", "minusassign");
    add (MMX_TIMESASSIGN, "*=", "timesassign");
    add (MMX_OVERASSIGN, "/=", "overassign");
    add (MMX_MODASSIGN, "%=", "modassign");
    add (MMX_LESSLESS, "<<", "lessless");
    add (MMX_GTRGTR, ">>", "gtrgtr");    
    add (MMX_EQUAL, "==", "equal");
    add (MMX_UNEQUAL, "!=", "unequal");
    add (MMX_LESSEQ, "<=", "lesseq");
    add (MMX_GTREQ, ">=", "gtreq");
    add (MMX_SEQAND, "&&", "seqand");
    add (MMX_SEQOR, "||", "seqor");
    add (MMX_PREINC, "++", "preinc");
    add (MMX_PREDEC, "--", "predec");
    add (MMX_ACCESS, "[]", "access");
    add (MMX_COMPLEMENT, "~", "complement");
    add (CCPP_ARROW, "->", "arrow");
  }
}

bool
is_opname (generic s) {
  return op_name->contains (s);
}

symbol
ccpp_back (generic s) {
  ccpp_init ();
  if (op_back->contains (s)) return as_symbol (op_back[s]);
  else if (is_symbol (s)) return as_symbol (s);
  return s;
}

symbol
ccpp_name (generic s) {
  ccpp_init ();
  if (op_name->contains (s)) return as_symbol (op_name[s]);
  else if (is_symbol (s)) return as_symbol (s);
  return s;
}

symbol
ccpp_opname (generic s) {
  ccpp_init ();
  if (op_name->contains (s))
    return symbol (string ("operator ") *
		   as_string (S (as_symbol (op_name[s]))));
  else return as_symbol (s);
}

symbol
ccpp_protect (generic s) {
  ccpp_init ();
  if (op_prot->contains (s)) return as_symbol (op_prot[s]);
  else return as_symbol (s);
}

/******************************************************************************
* The ccpp printer class
******************************************************************************/

class ccpp_printer_rep: public generic_rep {
private:
  ostream& out;
  list     sl;
  bool     lf_flag;
  bool     lf_done;

protected:
  size_t Size ();
  void Mark ();
  int Type ();
  void Print (ostream& out);

private:
  void pp_instr (generic g);
  void pp_init (generic g);
  void pp_prog (generic g);

  void pp_decl (generic var, bool large= true);
  void pp_decl (generic var, generic type, bool large);
  void pp_type (generic g);
  void pp_funarg (generic g);
  void pp_funargs (generic g);

  void pp_expr (generic g);
  void pp_exprs (generic g);
  void pp_assign (generic g);

  void pp_cond (generic g);
  void pp_seqor (generic g);
  void pp_seqand (generic g);
  void pp_or (generic g);
  void pp_xor (generic g);
  void pp_and (generic g);
  void pp_eq (generic g);
  void pp_rel (generic g);

  void pp_shift (generic g);
  void pp_lshift (generic g);
  void pp_rshift (generic g);
  void pp_add (generic g);
  void pp_mult (generic g);

  void pp_unary (generic g);
  void pp_postfix (generic g);
  void pp_primary (generic g);

public:
  inline ccpp_printer_rep (ostream& out2): out (out2) {}
  inline ~ccpp_printer_rep () {}

  void pp (generic g);

  friend class ccpp_printer;
};

class ccpp_printer {
DEFAULT_METHODS(ccpp_printer);
public:
  inline ccpp_printer (ostream& out):
    rep (new ccpp_printer_rep (out)) {}
};

/******************************************************************************
* Required routines for generic structures
******************************************************************************/

size_t
ccpp_printer_rep::Size () {
  return sizeof (ccpp_printer_rep);
}

void
ccpp_printer_rep::Mark () {
  mark (out);
  mark (sl);
  mark (lf_flag);
  mark (lf_done);
}

int
ccpp_printer_rep::Type () {
  return CCPP_PRINTER_TYPE;
}

void
ccpp_printer_rep::Print (ostream& out) {
  out << "#ccpp_printer<>";
}

/******************************************************************************
* Instructions
******************************************************************************/

void
ccpp_printer_rep::pp_instr (generic g) {
  if (is_func (g, CCPP_VERBATIM, 1) && is_string (g[1])) {
    string s= as_string (g[1]);
    int i, n=N(s);
    for (i=0; i<n; i++)
      if (s[i] == '\n') out << lf;
      else out << s[i];
  }
  else if (is_func (g, CCPP_NAMESPACE, 2)) {
    out << "namespace " << g[1] << " {" << lf << indent;
    pp_prog (g[2]);
    out << lf << unindent << "}" << lf;
  }
  else if (is_func (g, CCPP_INITIALIZE)) {
    pp_init (g);
    generic last= cAr (as_list (g));
    if (is_func (last, MMX_BEGIN)) pp_instr (last);
    else pp_instr (Begin (last));
  }
  else if (is_func (g, MMX_BEGIN)) {
    out << "{" << lf << indent;
    pp_prog (g);
    out << lf << unindent << "}" << lf;
  }
  else if (is_func (g, MMX_IF, 2)) {
    out << "if (";
    pp_expr (g[1]);
    out << ") ";
    if (is_func (g[2], MMX_IF)) pp_instr (Begin (g[2]));
    else pp_instr (g[2]);
  }
  else if (is_func (g, MMX_IF, 3)) {
    out << "if (";
    pp_expr (g[1]);
    out << ") ";
    if (is_func (g[2], MMX_IF, 2)) pp_instr (Begin (g[2]));
    else pp_instr (g[2]);
    out << "else ";
    pp_instr (g[3]);
  }
  else if (g == MMX_BREAK)
    out << "break;" << lf;
  else if (g == MMX_CONTINUE)
    out << "continue;" << lf;
  else if (is_func (g, MMX_RETURN, 1)) {
    out << "return ";
    pp_expr (g[1]);
    out << ";" << lf;
  }
  else pp_decl (g);
}

void
ccpp_printer_rep::pp_init (generic g) {
  list l= cDr (cdr (as_list (g)));
  bool first= true;
  while (!nil (l)) {
    generic init= car (l);
    if (first) out << ": ";
    else out << ", ";
    pp_expr (init);
    first= false;
    l= cdr (l);
  }
}

void
ccpp_printer_rep::pp_prog (generic g) {
  if (is_func (g, MMX_BEGIN) || is_func (g, CCPP_STRUCT)) {
    lf_flag= false;
    lf_done= true;
    list l= cdr (as_list (g));
    while (!nil (l)) {
      if (lf_flag) { out << lf; lf_flag= false; lf_done= true; }
      generic x= car (l);
      pp_instr (x);
      l= cdr (l);
      lf_done= false;
    }
  }
  else pp_instr (g);
}

/******************************************************************************
* Declarations
******************************************************************************/

void
ccpp_printer_rep::pp_decl (generic var, generic type, bool large) {
  // prefixes
  if (is_func (type, CCPP_CONST, 1)) {
    out << "const ";
    pp_decl (var, type[1], large);
  }
  else if (is_func (type, CCPP_EXTERN, 1)) {
    out << "extern ";
    pp_decl (var, type[1], large);
  }
  else if (is_func (type, CCPP_EXTERN, 2)) {
    out << "extern " << type[1] << " ";
    pp_decl (var, type[2], large);
  }
  else if (is_func (type, CCPP_INLINE, 1)) {
    out << "inline ";
    pp_decl (var, type[1], large);
  }
  else if (is_func (type, CCPP_STATIC, 1)) {
    out << "static ";
    pp_decl (var, type[1], large);
  }
  else if (is_func (type, CCPP_VIRTUAL, 1)) {
    out << "virtual ";
    pp_decl (var, type[1], large);
  }
  else if (is_func (type, MMX_FORALL, 2)) {
    generic x= type[2];
    out << "template<";
    pp_funarg (type[1]);
    while (is_func (x, MMX_FORALL, 2)) {
      out << ", ";
      pp_funarg (x[1]);
      x= x[2];
    }
    out << "> ";
    pp_decl (var, x, large);
  }

  // rewriting into C-form
  else if (is_func (type, MMX_ALIAS, 1))
    pp_decl (list (CCPP_UNALIAS, var), type[1], large);
  else if (is_func (type, MMX_POINTER, 1))
    pp_decl (list (CCPP_UNPOINTER, var), type[1], large);
  else if (is_func (type, MMX_ARRAY))
    pp_decl (cons (MMX_ACCESS, var, cddr (as_list (type))), type[1], large);
  else if (is_func (type, MMX_FUNCTION, 2)) {
    generic x= type[1];
    list args (x);
    if (is_func (x, MMX_CROSS)) args= cdr (as_list (x));
    if (x == MMX_VOID) args= list ();
    pp_decl (cons (CCPP_UNFUNCTION, var, args), type[2], large);
  }
  else if (is_func (type, CCPP_CONSTRUCTOR, 1)) {
    generic x= type[1];
    list args (x);
    if (is_func (x, MMX_CROSS)) args= cdr (as_list (x));
    if (x == MMX_VOID) args= list ();
    pp_decl (cons (CCPP_UNFUNCTION, var, args), symbol (""), large);
  }
  else if (is_func (type, CCPP_DESTRUCTOR, 0))
    pp_decl (list (CCPP_UNFUNCTION, var), symbol ("~"), large);
  else if (is_func (type, CCPP_INHERIT)) {
    out << "struct ";
    pp_type (var);
    out << ": ";
    list l= cdr (as_list (type));
    while (!nil (l)) {
      pp_type (car (l));
      l= cdr (l);
      if (!nil (l)) out << ", ";
    }
  }

  // normal declarations
  else {
    pp_type (type);
    if (type == symbol (""));
    else if (type == symbol ("~"));
    else if (large && (type != MMX_CLASS)) out << lf;
    else out << " ";
    pp_expr (var);
  }
}

void
ccpp_printer_rep::pp_decl (generic g, bool large) {
  if (is_func (g, CCPP_TYPEDEF, 2)) {
    out << "typedef ";
    pp_decl (list (MMX_TYPE, g[1], g[2]));
  }
  else if (is_func (g, MMX_DEFINE, 2) || is_func (g, MMX_ASSIGN, 2)) {
    bool large2= false;
    bool emptyb= false;
    if (is_func (g[1], MMX_TYPE, 2)) {
      large2= is_func (g[2], CCPP_INITIALIZE);
      large2 |= is_func (g[2], MMX_BEGIN);
      large2 |= is_func (g[2], CCPP_STRUCT);
      emptyb |= is_func (g[2], MMX_BEGIN, 0);
      emptyb |= is_func (g[2], CCPP_INITIALIZE) &&
	is_func (g[2][N(g[2])-1], MMX_BEGIN, 0);
      if (large && large2 && (!lf_done) && (!emptyb)) out << lf;
      if (is_func (g[2], CCPP_CONSTRUCT)) {
	generic var= cons (g[1][1], cdr (as_list (g[2])));
	if (is_func (g, MMX_DEFINE, 2))
	  pp_decl (var, list (CCPP_CONST, g[1][2]), large2);
	else pp_decl (var, g[1][2], large2);
      }
      else {
	if (is_func (g, MMX_DEFINE, 2) && (!large2))
	  pp_decl (g[1][1], list (CCPP_CONST, g[1][2]), large2);
	else pp_decl (g[1][1], g[1][2], large2);
      }
    }
    else pp_cond (g[1]);
    if (large && large2) {
      generic h= g[2];
      if (is_func (h, CCPP_INITIALIZE)) {
	pp_init (h);
	generic last= cAr (as_list (h));
	if (is_func (last, MMX_BEGIN)) h= last;
	else h= Begin (last);
      }
      if (emptyb) {
	out << " {";
	pp_prog (h);
	out << "}";
	out << lf;
	lf_flag= false;
      }
      else {
	out << " {" << lf << indent;
	pp_prog (h);
	out << unindent << "}";
	if (is_func (h, CCPP_STRUCT)) out << ";";
	out << lf;
	lf_flag= true;
      }
    }
    else {
      if (!is_func (g[2], CCPP_CONSTRUCT)) {
	out << " = ";
	pp_expr (g[2]);
      }
      if (large) out << ";" << lf;
    }
  }
  else if (is_func (g, MMX_TYPE, 2)) {
    pp_decl (g[1], g[2], false);
    if (large) out << ";" << lf;
  }
  else if (is_func (g, CCPP_USING, 1)) {
    out << "using ";
    pp_expr (g[1]);
    out << ";" << lf;
  }
  else {
    pp_expr (g);
    out << ";" << lf;
  }
}

/******************************************************************************
* Function arguments in function declarations
******************************************************************************/

static void
get_template_syntax (generic& fun, char& left, char& right) {
  left= '<'; right= '>';
  if (is_symbol (fun) &&
      is_string (S (as_symbol (fun))) &&
      starts (as_string (S (as_symbol (fun))), "macro ")) {
    string s= as_string (S (as_symbol (fun)));
    fun= symbol (s (6, N(s)));
    left= '('; right= ')';
  }
}

void
ccpp_printer_rep::pp_type (generic g) {
  if (is_func (g, MMX_CROSS, 0)) out << "void";
  else if (is_func (g, MMX_CROSS, 1)) pp_type (g[1]);
  else if (is_func (g, MMX_ALIAS, 1)) {
    pp_type (g[1]);
    out << "&";
  }
  else if (is_func (g, MMX_POINTER, 1)) {
    pp_type (g[1]);
    out << "*";
  }
  else if (is_func (g, CCPP_SCOPE, 2)) {
    pp_type (g[1]);
    out << "::";
    pp_type (g[2]);
  }
  else if (is_func (g, CCPP_CONST, 1)) {
    out << "const ";
    pp_type (g[1]);
  }
  else if (is_func (g, CCPP_PUBLIC, 1)) {
    out << "public ";
    pp_type (g[1]);
  }
  else if (is_list (g)) {
    list l= as_list (g);
    char left, right;
    generic fun= car (l);
    get_template_syntax (fun, left, right);
    pp_type (fun);
    l= cdr (l);
    out << left;
    while (!nil (l)) {
      pp_type (car (l));
      if (nil (cdr (l)) && is_list (car (l)))
	out << " ";
      l= cdr (l);
      if (!nil (l)) out << ", ";
    }
    out << right;
  }
  else if (g == MMX_CLASS) out << "struct";
  else if (g == MMX_VOID) out << "void";
  else if (g == MMX_BOOL) out << "bool";
  else if (g == MMX_INT) out << "int";
  else if (g == MMX_STRING) out << "mmx::string";
  else if (g == MMX_OSTREAM) out << "ostream";
  else if (g == MMX_LITERAL_INT) out << "string";
  else if (g == MMX_LITERAL_FLOAT) out << "string";
  else if (g == MMX_LITERAL_STRING) out << "string";
  else out << g;
}

void
ccpp_printer_rep::pp_funarg (generic g) {
  if (is_func (g, MMX_TYPE, 2) && (g[2] == MMX_CLASS))
    pp_decl (g[1], symbol ("class"), false);
  else if (is_func (g, MMX_TYPE, 2))
    pp_decl (g[1], g[2], false);
  else if (is_func (g, MMX_DEFINE, 2) || is_func (g, MMX_ASSIGN, 2))
    pp_decl (g, false);
  else pp_type (g);
}

void
ccpp_printer_rep::pp_funargs (generic g) {
  if (N(g) == 1) pp_funarg (car (g));
  if (N(g) <= 1) return;
  pp_funarg (car (g));
  out << ", ";
  pp_funargs (cdr (g));
}

/******************************************************************************
* Expressions
******************************************************************************/

void
ccpp_printer_rep::pp_expr (generic g) {
  pp_assign (g);
}

void
ccpp_printer_rep::pp_exprs (generic g) {
  if (N(g) == 1) pp_expr (car (g));
  if (N(g) <= 1) return;
  pp_expr (car (g));
  out << ", ";
  pp_exprs (cdr (g));
}

/******************************************************************************
* Assignments
******************************************************************************/

void
ccpp_printer_rep::pp_assign (generic g) {
  if (is_func (g, MMX_ASSIGN, 2)) {
    pp_cond (g[1]);
    out << " = ";
    pp_assign (g[2]);
  }
  else if (is_func (g, MMX_PLUSASSIGN, 2)) {
    pp_cond (g[1]);
    out << " += ";
    pp_assign (g[2]);
  }
  else if (is_func (g, MMX_MINUSASSIGN, 2)) {
    pp_cond (g[1]);
    out << " -= ";
    pp_assign (g[2]);
  }
  else if (is_func (g, MMX_TIMESASSIGN, 2)) {
    pp_cond (g[1]);
    out << " *= ";
    pp_assign (g[2]);
  }
  else if (is_func (g, MMX_OVERASSIGN, 2)) {
    pp_cond (g[1]);
    out << " /= ";
    pp_assign (g[2]);
  }
  else if (is_func (g, MMX_DIVASSIGN, 2)) {
    pp_cond (g[1]);
    out << " /= ";
    pp_assign (g[2]);
  }
  else if (is_func (g, MMX_MODASSIGN, 2)) {
    pp_cond (g[1]);
    out << " %= ";
    pp_assign (g[2]);
  }
  else pp_cond (g);
}

/******************************************************************************
* Conditional expressions
******************************************************************************/

void
ccpp_printer_rep::pp_cond (generic g) {
  if (is_func (g, MMX_IF, 3)) {
    pp_expr (g[1]);
    out << " ? ";
    pp_expr (g[2]);
    out << " : ";
    pp_expr (g[3]);
  }
  else pp_seqor (g);
}

/******************************************************************************
* Logical expressions
******************************************************************************/

void
ccpp_printer_rep::pp_seqor (generic g) {
  if (is_func (g, MMX_SEQOR, 2)) {
    pp_seqor (g[1]);
    out << " || ";
    pp_seqor (g[2]);
  }
  else pp_seqand (g);
}

void
ccpp_printer_rep::pp_seqand (generic g) {
  if (is_func (g, MMX_SEQAND, 2)) {
    pp_seqand (g[1]);
    out << " && ";
    pp_seqand (g[2]);
  }
  else pp_or (g);
}

void
ccpp_printer_rep::pp_or (generic g) {
  if (is_func (g, MMX_IMPLIES, 2))
    pp_or (list (MMX_OR, list (MMX_NOT, g[1]), g[2]));

  else if (is_func (g, MMX_OR, 2)) {
    pp_or (g[1]);
    out << " | ";
    pp_or (g[2]);
  }
  else pp_xor (g);
}

void
ccpp_printer_rep::pp_xor (generic g) {
  if (is_func (g, MMX_XOR, 2)) {
    pp_xor (g[1]);
    out << " ^ ";
    pp_xor (g[2]);
  }
  else pp_and (g);
}

void
ccpp_printer_rep::pp_and (generic g) {
  if (is_func (g, MMX_AND, 2)) {
    pp_and (g[1]);
    out << " & ";
    pp_and (g[2]);
  }
  else pp_eq (g);
}

/******************************************************************************
* Relational expressions
******************************************************************************/

void
ccpp_printer_rep::pp_eq (generic g) {
  if (is_func (g, MMX_EQUIV, 2))
    pp_eq (list (MMX_EQUAL, g[1], g[2]));

  else if (is_func (g, MMX_EQUAL, 2)) {
    pp_rel (g[1]);
    out << " == ";
    pp_rel (g[2]);
  }
  else if (is_func (g, MMX_UNEQUAL, 2)) {
    pp_rel (g[1]);
    out << " != ";
    pp_rel (g[2]);
  }
  else pp_rel (g);
}

void
ccpp_printer_rep::pp_rel (generic g) {
  if (is_func (g, MMX_LESS, 2)) {
    pp_shift (g[1]);
    out << " < ";
    pp_shift (g[2]);
  }
  else if (is_func (g, MMX_LESSEQ, 2)) {
    pp_shift (g[1]);
    out << " <= ";
    pp_shift (g[2]);
  }
  else if (is_func (g, MMX_GTR, 2)) {
    pp_shift (g[1]);
    out << " > ";
    pp_shift (g[2]);
  }
  else if (is_func (g, MMX_GTREQ, 2)) {
    pp_shift (g[1]);
    out << " >= ";
    pp_shift (g[2]);
  }
  else pp_shift (g);
}

/******************************************************************************
* Arithmetic expressions
******************************************************************************/

void
ccpp_printer_rep::pp_shift (generic g) {
  if (is_func (g, MMX_LESSLESS, 2)) {
    pp_lshift (g[1]);
    out << " << ";
    pp_add (g[2]);
  }
  else if (is_func (g, MMX_GTRGTR, 2)) {
    pp_add (g[1]);
    out << " >> ";
    pp_rshift (g[2]);
  }
  else pp_add (g);
}

void
ccpp_printer_rep::pp_lshift (generic g) {
  if (is_func (g, MMX_LESSLESS, 2)) {
    pp_lshift (g[1]);
    out << " << ";
    pp_add (g[2]);
  }
  else pp_add (g);
}

void
ccpp_printer_rep::pp_rshift (generic g) {
  if (is_func (g, MMX_GTRGTR, 2)) {
    pp_add (g[1]);
    out << " >> ";
    pp_rshift (g[2]);
  }
  else pp_add (g);
}

void
ccpp_printer_rep::pp_add (generic g) {
  if (is_func (g, MMX_PLUS, 2)) {
    pp_add (g[1]);
    out << " + ";
    pp_add (g[2]);
  }
  else if (is_func (g, MMX_MINUS, 2)) {
    pp_add (g[1]);
    out << " - ";
    pp_mult (g[2]);
  }
  else pp_mult (g);
}

void
ccpp_printer_rep::pp_mult (generic g) {
  if (is_func (g, MMX_TIMES, 2)) {
    pp_mult (g[1]);
    out << " * ";
    pp_mult (g[2]);
  }
  else if (is_func (g, MMX_OVER, 2)) {
    pp_mult (g[1]);
    out << " / ";
    pp_unary (g[2]);
  }
  else if (is_func (g, MMX_DIV, 2)) {
    pp_mult (g[1]);
    out << " / ";
    pp_unary (g[2]);
  }
  else if (is_func (g, MMX_MOD, 2)) {
    pp_mult (g[1]);
    out << " % ";
    pp_unary (g[2]);
  }
  else pp_unary (g);
}

/******************************************************************************
* Unary, postfix and primary expressions
******************************************************************************/

void
ccpp_printer_rep::pp_unary (generic g) {
  if (is_func (g, MMX_NOTLESS, 2))
    pp_unary (list (MMX_NOT, list (MMX_LESS, g[1], g[2])));
  else if (is_func (g, MMX_NOTLESSEQ, 2))
    pp_unary (list (MMX_NOT, list (MMX_LESSEQ, g[1], g[2])));
  else if (is_func (g, MMX_NOTGTR, 2))
    pp_unary (list (MMX_NOT, list (MMX_GTR, g[1], g[2])));
  else if (is_func (g, MMX_NOTGTREQ, 2))
    pp_unary (list (MMX_NOT, list (MMX_GTREQ, g[1], g[2])));

  else if (is_func (g, MMX_PREINC, 1)) {
    out << "++";
    pp_unary (g[1]);
  }
  else if (is_func (g, MMX_PREDEC, 1)) {
    out << "--";
    pp_unary (g[1]);
  }
  else if (is_func (g, CCPP_UNPOINTER, 1)) {
    out << "*";
    pp_unary (g[1]);
  }
  else if (is_func (g, CCPP_UNALIAS, 1)) {
    out << "&";
    pp_unary (g[1]);
  }
  else if (is_func (g, MMX_PLUS, 1)) {
    out << "+";
    pp_unary (g[1]);
  }
  else if (is_func (g, MMX_MINUS, 1)) {
    out << "-";
    pp_unary (g[1]);
  }
  else if (is_func (g, MMX_NOT, 1)) {
    out << "!";
    pp_unary (g[1]);
  }
  else if (is_func (g, MMX_COMPLEMENT, 1)) {
    out << "~";
    pp_unary (g[1]);
  }
  else if (is_func (g, MMX_TRANSTYPE, 2)) {
    out << "(";
    pp_type (g[2]);
    out << ") ";
    pp_unary (g[1]);
  }
  else if (is_func (g, CCPP_CAST, 2)) {
    pp_type (list (symbol ("reinterpret_cast"), g[2])); // hack to produce <>
    out << " (";
    pp_expr (g[1]);
    out << ")";
  }
  else if (is_func (g, CCPP_DELETE, 1)) {
    out << "delete ";
    pp_expr (g[1]);
  }
  else if (is_func (g, CCPP_NEW)) {
    out << "new ";
    pp_type (g[1]);
    out << " (";
    pp_exprs (cdr (cdr (g)));
    out << ")";
  }
  else pp_postfix (g);
}

/******************************************************************************
* Postfix expressions
******************************************************************************/

static bool
is_applicable (generic g) {
  if (N(g)>0) return true;
  if (is_literal (g)) return true;
  if (gentype (g) == SYMBOL_TYPE)
    return is_applicable (S(g));
  if (gentype (g) == STRING_TYPE) {
    string s= as_string (g);
    if (N(s)==0) return false;
    int i;
    for (i=0; i<N(s); i++)
      if (((s[i]<'0') || (s[i]>'9')) &&
          ((s[i]<'a') || (s[i]>'z')) &&
          ((s[i]<'A') || (s[i]>'Z')) &&
          ((s[i]<'_') || (s[i]>'_')))
        return false;
    return true;
  }
  return false;
}

void
ccpp_printer_rep::pp_postfix (generic g) {
  if (is_func (g, MMX_FUNCTION, 2))
    pp_postfix (list (symbol ("into"), g[1], g[2]));
  else if (is_func (g, MMX_DOTS, 2))
    pp_postfix (list (symbol ("ldots"), g[1], g[2]));
  else if (is_func (g, MMX_OPLUS, 2))
    pp_postfix (list (symbol ("oplus"), g[1], g[2]));
  else if (is_func (g, MMX_OMINUS, 2))
    pp_postfix (list (symbol ("ominus"), g[1], g[2]));
  else if (is_func (g, MMX_OTIMES, 2))
    pp_postfix (list (symbol ("otimes"), g[1], g[2]));
  else if (is_func (g, MMX_OOVER, 2))
    pp_postfix (list (symbol ("oover"), g[1], g[2]));
  else if (is_func (g, MMX_COMPOSE, 2))
    pp_postfix (list (symbol ("compose"), g[1], g[2]));
  if (is_func (g, MMX_NUMBER, 1))
    pp_postfix (list (symbol ("number"), g[1]));
  else if (is_func (g, MMX_DOLLAR, 1))
    pp_postfix (list (symbol ("dollar"), g[1]));
  else if (is_func (g, MMX_POWER, 2))
    pp_postfix (list (symbol ("pow"), g[1], g[2]));
  else if (is_func (g, MMX_FACTORIAL, 1))
    pp_postfix (list (symbol ("fact"), g[1]));
  else if (is_func (g, MMX_PRIME, 1))
    pp_postfix (list (symbol ("prime"), g[1]));

  else if (is_func (g, CCPP_SCOPE, 2)) {
    pp_type (g[1]);
    out << "::";
    pp_postfix (g[2]);
  }
  else if (is_func (g, MMX_DOT, 2)) {
    pp_postfix (g[1]);
    out << ".";
    pp_primary (g[2]);
  }
  else if (is_func (g, CCPP_ARROW, 2)) {
    pp_postfix (g[1]);
    out << "->";
    pp_primary (g[2]);
  }
  else if (is_func (g, MMX_POSTINC, 1)) {
    pp_postfix (g[1]);
    out << "++";
  }
  else if (is_func (g, MMX_POSTDEC, 1)) {
    pp_postfix (g[1]);
    out << "--";
  }
  else if (is_func (g, CCPP_UNFUNCTION)) {
    pp_postfix (g[1]);
    out << " (";
    pp_funargs (cdr (cdr (g)));
    out << ")";
  }
  else if (is_func (g, CCPP_SPECIALIZE)) {
    char left, right;
    generic fun= g[1];
    get_template_syntax (fun, left, right);
    pp_postfix (fun);
    out << left;
    pp_funargs (cdr (cdr (g)));
    if (!is_symbol (cAr (as_list (g)))) out << " ";
    out << right;
  }
  else if (is_func (g, MMX_ACCESS)) {
    pp_postfix (g[1]);
    out << "[";
    pp_exprs (cdr (cdr (g)));
    out << "]";
  }
  else if (((N(g) > 0) && is_applicable (g[0])) ||
	   ((!nil (sl)) && (car (sl) == g))) {
    pp_postfix  (car (g));
    out << " (";
    pp_exprs (cdr (g));
    out << ")";
  }
  else pp_primary (g);
}

/******************************************************************************
* Primary expressions
******************************************************************************/

void
ccpp_printer_rep::pp_primary (generic g) {
  if (is_opname (g)) out << ccpp_opname (g);
  else if (is_string (g)) out << quote (as_string (g));
  else if (N(g) == 0) out << g;
  else if (is_func (g, MMX_SQTUPLE) || is_func (g, MMX_TUPLE)) {
    if (N(g) == 0) out << "{}";
    if (N(g) <= 3) {
      out << "{ ";
      pp_exprs (cdr (g));
      out << " }";
    }
    else {
      list l= cdr (as_list (g));
      out << "{" << lf << indent;
      while (!nil (l)) {
	pp_expr (car (l));
	l= cdr (l);
	if (!nil (l)) out << "," << lf;
      }
      out << lf << unindent << "}";
    }
  }
  else if (is_func (g, CCPP_BRACKETS, 1)) {
    out << "(";
    pp_expr (g[1]);
    out << ")";
  }
  else {
    out << "(";
    sl= cons (g, sl);
    pp_expr (g);
    sl= cdr (sl);
    out << ")";
  }
}

/******************************************************************************
* Interface
******************************************************************************/

void
ccpp_printer_rep::pp (generic g) {
  out << blank;
  pp_prog (g);
}

void
ccpp_print (ostream& out, generic g) {
  serializer sout;
  ccpp_printer pp (sout->out);
  pp->pp (g);
  out << delsimpl (as_string (sout));
}

} // namespace mmx
