
/******************************************************************************
* MODULE     : ccpp.hh
* DESCRIPTION: Convert mathemagix trees to C or C++
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __CCPP_HH
#define __CCPP_HH
#include <symbols.hh>
namespace mmx {

void ccpp_print (ostream& out, generic g);

extern symbol CCPP_ARROW;
extern symbol CCPP_CAST;
extern symbol CCPP_CONSTRUCT;
extern symbol CCPP_CONSTRUCTOR;
extern symbol CCPP_DELETE;
extern symbol CCPP_DESTRUCTOR;
extern symbol CCPP_INHERIT;
extern symbol CCPP_INITIALIZE;
extern symbol CCPP_NAMESPACE;
extern symbol CCPP_NEW;
extern symbol CCPP_SCOPE;
extern symbol CCPP_SPECIALIZE;
extern symbol CCPP_STRUCT;
extern symbol CCPP_TYPEDEF;
extern symbol CCPP_UNALIAS;
extern symbol CCPP_UNPOINTER;
extern symbol CCPP_UNFUNCTION;
extern symbol CCPP_USING;
extern symbol CCPP_VERBATIM;

extern symbol CCPP_CONST;
extern symbol CCPP_EXTERN;
extern symbol CCPP_INLINE;
extern symbol CCPP_STATIC;
extern symbol CCPP_VIRTUAL;
extern symbol CCPP_PUBLIC;

extern symbol CCPP_BRACKETS;

symbol ccpp_back (generic s);
symbol ccpp_name (generic s);
symbol ccpp_opname (generic s);
symbol ccpp_protect (generic s);

inline generic Arrow (generic g1, generic g2) {
  return list (CCPP_ARROW, g1, g2); }
inline generic Cast (generic g1, generic g2) {
  return list (CCPP_CAST, g1, g2); }
inline generic Construct (generic g1, generic g2, generic g3) {
  return list (CCPP_CONSTRUCT, g1, g2, g3); }
inline generic Constructor () {
  return list (CCPP_CONSTRUCTOR, MMX_VOID); }
inline generic Constructor (generic g1) {
  return list (CCPP_CONSTRUCTOR, g1); }
inline generic Constructor (list l) {
  return list (CCPP_CONSTRUCTOR, Cross_var (l)); }
inline generic Delete (generic g) {
  return list (CCPP_DELETE, g); }
inline generic Extern (generic g) {
  return list (CCPP_EXTERN, g); }
inline generic Destructor () {
  return list (CCPP_DESTRUCTOR); }
inline generic Initialize (generic g1, generic g2) {
  return list (CCPP_INITIALIZE, g1, g2); }
inline generic Initialize (list l, generic g) {
  return list (CCPP_INITIALIZE) * l * list (g); }
inline generic Namespace (generic g1, generic g2) {
  return list (CCPP_NAMESPACE, g1, g2); }
inline generic New (generic g) {
  return list (CCPP_NEW, g); }
inline generic New (generic g1, generic g2) {
  return list (CCPP_NEW, g1, g2); }
inline generic New (generic g, list l) {
  return cons (CCPP_NEW, g, l); }
inline generic Scope (generic g1, generic g2) {
  return list (CCPP_SCOPE, g1, g2); }
inline generic Specialize (generic g1, generic g2) {
  return list (CCPP_SPECIALIZE, g1, g2); }
inline generic Specialize (generic g1, generic g2, generic g3) {
  return list (CCPP_SPECIALIZE, g1, g2, g3); }
inline generic Specialize (generic g1, generic g2, generic g3, generic g4) {
  return list (CCPP_SPECIALIZE, g1, g2, g3, g4); }
inline generic Struct (generic g1) {
  return list (CCPP_STRUCT, g1); }
inline generic Typedef (generic g1, generic g2) {
  return list (CCPP_TYPEDEF, g1, g2); }
inline generic Unalias (generic g1) {
  return list (CCPP_UNALIAS, g1); }
inline generic Unpointer (generic g1) {
  return list (CCPP_UNPOINTER, g1); }
inline generic Using (generic g1) {
  return list (CCPP_USING, g1); }

inline generic Const (generic g1) {
  return list (CCPP_CONST, g1); }
inline generic Inline (generic g1) {
  return list (CCPP_INLINE, g1); }
inline generic Public (generic g1) {
  return list (CCPP_PUBLIC, g1); }
inline generic Static (generic g1) {
  return list (CCPP_STATIC, g1); }
inline generic Virtual (generic g1) {
  return list (CCPP_VIRTUAL, g1); }

inline generic Brackets (generic g1) {
  return list (CCPP_BRACKETS, g1); }
inline generic Cast_var (generic g1, generic g2) {
  return Unpointer (Cast (Unalias (g1), Pointer (g2))); }

}
#endif // __CCPP_HH
