
/******************************************************************************
* MODULE     : function.cc
* DESCRIPTION: functions
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <function.hh>
#include <symbol.hh>
namespace mmx {

/******************************************************************************
* Required functions for generic structures
******************************************************************************/

size_t
function_rep::Size () {
  return sizeof (function_rep);
}

void
function_rep::Mark () {
  mark (name);
  mark (fn);
  mark (args);
}

int
function_rep::Type () {
  return FUNCTION_TYPE;
}

int
function_rep::Hash () {
  int h= hash (args);
  return (h<<11) + (h>>21) + ((int) fn);
}

bool
function_rep::Equal (generic g) {
  if (gentype (g) != FUNCTION_TYPE) return false;
  return function (this) == as_function (g);
}

generic
function_rep::Copy () {
  return copy (function (this));
}

int
function_rep::Arity () {
  return 0;
}

void
function_rep::Print (ostream& out) {
  out << function (this);
}

/******************************************************************************
* Routines for functions
******************************************************************************/

ostream&
operator << (ostream& out, function f) {
  if (nil (f->args)) return print (out, "#function", f->name);
  else return print (out, "#function", f->name, f->args);
}

function
copy (function f) {
  return function (f->name, f->fn, f->args);
}

function
add_arguments (function f, list args) {
  return function (f->name, f->fn, f->args * args);
}

} // namespace mmx
