
/******************************************************************************
* MODULE     : function.hh
* DESCRIPTION: generic functions
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __FUNCTION_HH
#define __FUNCTION_HH
#include <string.hh>
#include <list.hh>
namespace mmx {

class function;
class function_rep: public generic_rep {
public:
  generic      name;   // a descriptive name
protected:
  mmx_function fn;     // the function
  list         args;   // extra arguments

protected:
  size_t Size ();
  void Mark ();
  int Type ();
  int Hash ();
  bool Equal (generic g);
  generic Copy ();
  int Arity ();
  void Print (ostream& out);

public:
  inline function_rep (generic name2, mmx_function fn2, list args2):
    name (name2), fn (fn2), args (args2) {}

  friend class function;
  friend function copy (function t);
  friend function add_arguments (function t, list args);
  friend ostream& operator << (ostream& out, function a);
};

class function {
DEFAULT_METHODS(function);
public:
  inline function (mmx_function fn):
    rep (new function_rep (string ("function"), fn, list ())) {}
  inline function (string name, mmx_function fn):
    rep (new function_rep (name, fn, list ())) {}
  inline function (generic name, mmx_function fn):
    rep (new function_rep (name, fn, list ())) {}
  inline function (string name, mmx_function fn, list args):
    rep (new function_rep (name, fn, args)) {}
  inline function (generic name, mmx_function fn, list args):
    rep (new function_rep (name, fn, args)) {}

  inline generic operator () (generic g) {
    if (nil (rep->args)) return rep->fn (g);
    else return rep->fn (cons (g, rep->args)); }
  inline bool operator == (function f) {
    return (rep->fn == f->fn) && (rep->args == f->args); }
  inline bool operator != (function f) {
    return (rep->fn != f->fn) || (rep->args != f->args); }
  friend function copy (function a);
  friend ostream& operator << (ostream& out, function a);
};

}
#endif // __FUNCTION_HH
