
/******************************************************************************
* MODULE     : generic.cc
* DESCRIPTION: Abstract types
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <generic.hh>
#include <list.hh>
#include <table.hh>
#include <single.hh>
#include <pointer.hh>
#include <symbol.hh>
#include <mirror.hh>
#include <literal.hh>
#include <genalias.hh>
#include <value.hh>
namespace mmx {

GC_GLOBAL (table, source_location, table ());
GC_GLOBAL (print_format, current_print_format, normal);
GC_GLOBAL (int, current_indentation, 0);

GC_GLOBAL_STATIC_CHAR_ARRAY (indent_s, "| ");
GC_GLOBAL_STATIC_CHAR_ARRAY (unindent_s, "\b\b");
GC_GLOBAL_STATIC_CHAR_ARRAY (hrule_s, "-----------------------------------------------------------------------------\n");

extern int cycle_depth;
static void locate_sub (generic g, int& begin, int& end);

/******************************************************************************
* Basic routines
******************************************************************************/

int
generic_rep::Type () {
  return UNKNOWN_TYPE;
}

int
generic_rep::Hash () {
  return ((int) this) >> 2;
}

bool
generic_rep::Equal (generic g) {
  return ((void*) this) == ((void*) g.rep);
}

generic
generic_rep::Copy () {
  return generic (this);
}

int
generic_rep::Arity () {
  return 0;
}

generic&
generic_rep::Access (int i) {
  cout << lf << "g= " << generic (this) << lf;
  fatal_error ("no data access function associated to generic",
	       "generic_rep::Access", "generic.cc");
  exit (1);
}

generic
generic_rep::Range (int begin, int end) {
  cout << lf << "g= " << generic (this) << lf;
  fatal_error ("no range function associated to generic",
	       "generic_rep::Range", "generic.cc");
  exit (1);
}

generic&
generic_rep::Car () {
  cout << lf << "g= " << generic (this) << lf;
  fatal_error ("no data access function associated to generic",
	       "generic_rep::Car", "generic.cc");
  exit (1);
}

generic
generic_rep::Cdr () {
  cout << lf << "g= " << generic (this) << lf;
  fatal_error ("no range function associated to generic",
	       "generic_rep::Cdr", "generic.cc");
  exit (1);
}

void
generic_rep::Print (ostream& out) {
  out << "#unknown";
}

bool
generic_rep::Is_genalias () {
  return false;
}

bool
generic_rep::Is_gentable () {
  return false;
}

bool
generic_rep::Is_literal () {
  return false;
}

void
generic_rep::Assign_me (generic g) {
  cout << lf << "g= " << generic (this) << lf;
  fatal_error ("no assign function associated to generic",
	       "generic_rep::Assign", "generic.cc");
  exit (1);
}

/******************************************************************************
* Printing format
******************************************************************************/

ostream&
operator << (ostream& out, print_format x) {
  int i;
  switch (x) {
  case blank:
    indent_s[0]= ' ';
    indent_s[1]= ' ';
    break;
  case stroke:
    indent_s[0]= '|';
    indent_s[1]= ' ';
    break;
  case indent:
    current_indentation++;
    out << indent_s;
    break;
  case unindent:
    current_indentation--;
    out << unindent_s;
    break;
  case lf:
    out << "\n";
    for (i=0; i<current_indentation; i++)
      out << indent_s;
    break;
  case hrule:
    out << hrule_s;
    break;
  case expand_cycles:
    cycle_depth++;
    break;
  case compress_cycles:
    cycle_depth--;
    break;
  case normal:
  case condensed:
  case detailed:
  case indented:
    current_print_format= x;
    break;
  }
  return out;
}

bool
long_print_format () {
  return current_print_format == detailed;
}

ostream&
print (ostream& out, char* s, generic g1) {
  if (current_print_format == indented) {
    out << s;
    current_indentation++;
    out << lf << g1;
    current_indentation--;
    return out;
  }
  else return out << s << "<" << g1 << ">";
}

ostream&
print (ostream& out, char* s, generic g1, generic g2) {
  if (current_print_format == indented) {
    out << s;
    current_indentation++;
    out << lf << g1;
    out << lf << g2;
    current_indentation--;
    return out;
  }
  else return out << s << "<" << g1 << " " << g2 << ">";
}

/******************************************************************************
* Locations in source code
******************************************************************************/

void
generic_rep::index (int begin, int end) {
#ifdef SOURCE_TRACKING
  (void) begin; (void) end;
  pointer address ((void*) this);
  source_location (address)= list (single (begin), single (end));
#else
  (void) begin; (void) end;
#endif
}

void
generic_rep::index (generic g) {
#ifdef SOURCE_TRACKING
  (void) g;
  int begin, end;
  locate (g, begin, end);
  index (begin, end);
#else
  (void) g;
#endif
}

generic
extend (generic g, generic with) {
#ifdef SOURCE_TRACKING
  if (nil (g) || nil (with)) return g;
  int begin=  0x40000000;
  int end  = -0x40000000;
  locate_sub (g, begin, end);
  if ((begin != 0x40000000) && (end != -0x40000000))
    g -> index (begin, end);
  return g;
#else
  (void) with;
  return g;
#endif
}

static void
locate_sub (generic g, int& begin, int& end) {
  if (nil (g)) return;
  pointer address ((void*) g.rep);
  if (source_location -> contains (address)) {
    generic location= source_location [address];
    begin= min (begin, as_int (location[0]));
    end  = max (end  , as_int (location[1]));
  }
  else switch (gentype (g)) {
  case SYMBOL_TYPE:
    locate_sub (S (g), begin, end);
    break;
  case MIRROR_TYPE:
    locate_sub (S (as_mirror (g)), begin, end);
    break;
  case ALIAS_TYPE:
  case ALIASES_TYPE:
    locate_sub (as_genalias (g) -> get (), begin, end);
    break;
  case LIST_TYPE:
    locate_sub (car (g), begin, end);
    locate_sub (cdr (g), begin, end);
    break;
  case VALUE_TYPE:
    locate_sub (V (g), begin, end);
    break;
  case LITERAL_INT_TYPE:
  case LITERAL_FLOAT_TYPE:
  case LITERAL_STRING_TYPE:
    locate_sub (S (as_literal (g)), begin, end);
    break;
  }
}

void
locate (generic g, int& begin, int& end) {
#ifdef SOURCE_TRACKING
  begin=  0x40000000;
  end  = -0x40000000;
  locate_sub (g, begin, end);
  if (begin ==  0x40000000) begin= -1;
  if (end   == -0x40000000) end  = -1;
#else
  (void) g;
  begin= end= -1;
#endif
}

int
generic_rep::begin () {
  int _begin, _end;
  locate (this, _begin, _end);
  return _begin;
}

int
generic_rep::end () {
  int _begin, _end;
  locate (this, _begin, _end);
  return _end;
}

/******************************************************************************
* Other routines
******************************************************************************/

generic&
S (generic g) {
  if (gentype (g) == SYMBOL_TYPE) return S (as_symbol (g));
  type_mismatch (g, "symbol", "S", "generic.cc");
  exit (1);
}

void
fatal_error (char* message, char* routine, char* file) {
  cerr << "\nFatal error: " << message
       << "\n             in '" << routine
       << "' in file '" << file << "'\n";
  assert (false);
}

void
type_mismatch (generic g, char* expected_type) {
  cerr << "\nFatal error: expected a value of type " << expected_type
       << "\n             in 'as_" << expected_type
       << "' in file '" << expected_type << ".hh'"
       << "\nValue was  : " << g
       << "\nType was   : " << gentype (g) << "\n";
  assert (false);
}

void
type_mismatch (generic g, char* expected_type, char* routine, char* file) {
  cerr << "\nFatal error: expected a value of type " << expected_type
       << "\n             in '" << routine
       << "' in file '" << file << "'"
       << "\nValue was  : " << g
       << "\nType was   : " << gentype (g) << "\n";
  assert (false);
}

} // namespace mmx
