
/******************************************************************************
* MODULE     : list.hh
* DESCRIPTION: lists
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __LIST_HH
#define __LIST_HH
#include <generic.hh>
namespace mmx {

/******************************************************************************
* The list class and its representation class
******************************************************************************/

class list_rep;
class list {
DEFAULT_METHODS_PROTO(list);
public:
  inline list ();
  inline list (generic g1);
  inline list (generic g1, generic g2);
  inline list (generic g1, generic g2, generic g3);
  inline list (generic g1, generic g2, generic g3, generic g4);
  inline list (generic g1, generic g2, generic g3, generic g4, generic g5);
  inline list (generic g1, generic g2, generic g3, generic g4,
	       generic g5, generic g6);

  friend int N (list l);
  friend bool operator == (list a, list b);
  friend bool operator != (list a, list b);
  friend list copy (list l);
  generic& operator [] (int i);
  list operator () (int start, int end);
  inline friend bool nil (list l);
  inline friend bool atom (list l);
  friend ostream& operator << (ostream& out, list a);

  inline friend generic& car (list l);
  inline friend generic& cadr (list l);
  inline friend generic& caddr (list l);
  inline friend generic& cadddr (list l);
  inline friend list& cdr (list l);
  inline friend list& cddr (list l);
  inline friend list& cdddr (list l);
  inline friend list& cddddr (list l);

  inline friend list cons (generic g1, list g2);
  inline friend list cons (generic g1, generic g2, list g3);
  inline friend list cons (generic g1, generic g2, generic g3, list g4);
  inline friend list cons (generic g1, generic g2, generic g3,
			   generic g4, list g5);
  inline friend list cons (generic g1, generic g2, generic g3,
			   generic g4, generic g5, list g6);

  inline friend list cons (generic g1, generic g2);
  inline friend list cons (generic g1, generic g2, generic g3);
  inline friend list cons (generic g1, generic g2, generic g3, generic g4);
  inline friend list cons (generic g1, generic g2, generic g3,
			   generic g4, generic g5);
  inline friend list cons (generic g1, generic g2, generic g3,
			   generic g4, generic g5, generic g6);
};

class list;
class list_rep: public generic_rep {
protected:
  generic item;
  list    next;

protected:
  size_t Size ();
  void Mark ();
  int Type ();
  int Hash ();
  bool Equal (generic x);
  generic Copy ();
  int Arity ();
  generic& Access (int i);
  generic Range (int begin, int end);
  generic& Car ();
  generic Cdr ();
  void Print (ostream& out);

public:
  inline list_rep (generic g);
  inline list_rep (generic g, list l);
  inline ~list_rep ();

  friend class list;
  friend generic& car (list l);
  friend generic& cadr (list l);
  friend generic& caddr (list l);
  friend generic& cadddr (list l);
  friend list&    cdr (list l);
  friend list&    cddr (list l);
  friend list&    cdddr (list l);
  friend list&    cddddr (list l);
  friend bool     atom (list l);
};

/******************************************************************************
* Inline routines
******************************************************************************/

inline list_rep::list_rep (generic g): item (g) {}
inline list_rep::list_rep (generic g, list l): item (g), next (l) {}
inline list_rep::~list_rep () {}
inline generic cAr (list l) { return l[N(l)-1]; }
inline list cDr (list l) { return l(0,N(l)-1); }

DEFAULT_METHODS_CODE(list);

inline list::list ():
  rep (NULL) {}
inline list::list (generic g1):
  rep (new list_rep (g1)) {}
inline list::list (generic g1, generic g2):
  rep (new list_rep (g1, list (g2))) {}
inline list::list (generic g1, generic g2, generic g3):
  rep (new list_rep (g1, list (g2, g3))) {}
inline list::list (generic g1, generic g2, generic g3, generic g4):
  rep (new list_rep (g1, list (g2, g3, g4))) {}
inline list::list (generic g1, generic g2, generic g3, generic g4, generic g5):
  rep (new list_rep (g1, list (g2, g3, g4, g5))) {}
inline list::list (generic g1, generic g2, generic g3, generic g4,
	     generic g5, generic g6):
  rep (new list_rep (g1, list (g2, g3, g4, g5, g6))) {}

inline bool nil (list l) { return l.rep == NULL; }
inline bool atom (list l) {
  return (l.rep != NULL) && (l.rep->next.rep == NULL); }

inline generic& car (list l) {
  return l.rep->item; }
inline generic& cadr (list l) {
  return l.rep->next.rep->item; }
inline generic& caddr (list l) {
  return l.rep->next.rep->next.rep->item; }
inline generic& cadddr (list l) {
  return l.rep->next.rep->next.rep->next.rep->item; }
inline list& cdr (list l) {
  return l.rep->next; }
inline list& cddr (list l) {
  return l.rep->next.rep->next; }
inline list& cdddr (list l) {
  return l.rep->next.rep->next.rep->next; }
inline list& cddddr (list l) {
  return l.rep->next.rep->next.rep->next.rep->next; }

inline list cons (generic g1, list g2) {
  return list (new list_rep (g1, g2)); }
inline list cons (generic g1, generic g2, list g3) {
  return cons (g1, cons (g2, g3)); }
inline list cons (generic g1, generic g2, generic g3, list g4) {
  return cons (g1, cons (g2, cons (g3, g4))); }
inline list cons (generic g1, generic g2, generic g3, generic g4, list g5) {
  return cons (g1, cons (g2, cons (g3, cons (g4, g5)))); }
inline list cons (generic g1, generic g2, generic g3,
		  generic g4, generic g5, list g6) {
  return cons (g1, cons (g2, cons (g3, cons (g4, cons (g5, g6))))); }

inline list cons (generic g1, generic g2) {
  return cons (g1, as_list (g2)); }
inline list cons (generic g1, generic g2, generic g3) {
  return cons (g1, cons (g2, as_list (g3))); }
inline list cons (generic g1, generic g2, generic g3, generic g4) {
  return cons (g1, cons (g2, cons (g3, as_list (g4)))); }
inline list cons (generic g1, generic g2, generic g3, generic g4, generic g5) {
  return cons (g1, cons (g2, cons (g3, cons (g4, as_list (g5))))); }
inline list cons (generic g1, generic g2, generic g3,
		  generic g4, generic g5, generic g6) {
  return cons (g1, cons (g2, cons (g3,
    cons (g4, cons (g5, as_list (g6)))))); }

/******************************************************************************
* Further routines
******************************************************************************/

list reverse (list l);
list operator * (list l1, list l2);
list apply_all (generic g, list l);
bool is_list (generic g, generic g1);
bool is_list (generic g, generic g1, generic g2);

} // namespace mmx
#endif // __LIST_HH
