
/******************************************************************************
* MODULE     : literal.cc
* DESCRIPTION: Literal objects
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <literal.hh>
namespace mmx {

/******************************************************************************
* Required functions for generic structures
******************************************************************************/

size_t
literal_rep::Size () {
  return sizeof (literal_rep);
}

void
literal_rep::Mark () {
  mark (s);
  mark (type);
}

int
literal_rep::Type () {
  return type;
}

int
literal_rep::Hash () {
  return hash (s) ^ type;
}

bool
literal_rep::Equal (generic g) {
  if (gentype (g) != type) return false;
  return literal (this) == as_literal (g);
}

void
literal_rep::Print (ostream& out) {
  out << literal (this);
}

bool
literal_rep::Is_literal () {
  return true;
}

/******************************************************************************
* Printing literals
******************************************************************************/

ostream&
operator << (ostream& out, literal lit) {
  switch (lit->type) {
  case LITERAL_INT_TYPE:
    return out << "#literal_int<" << lit->s << ">";
  case LITERAL_FLOAT_TYPE:
    return out << "#literal_float<" << lit->s << ">";
  case LITERAL_STRING_TYPE:
    return out << "#literal_string<" << lit->s << ">";
  default:
    return out << "#literal_unknown<" << lit->s << ">";
  }
}

} // namespace mmx
