
/******************************************************************************
* MODULE     : literal.hh
* DESCRIPTION: Literal objects
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* Mathemagix uses literal integers, strings and floats
* These are really strings with a distinct type
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __LITERAL_HH
#define __LITERAL_HH
#include <symbol.hh>
namespace mmx {

class literal;
class literal_rep: public generic_rep {
protected:
  symbol s;
  int    type;

protected:
  size_t Size ();
  void Mark ();
  int Type ();
  int Hash ();
  bool Equal (generic x);
  void Print (ostream& out);
  bool Is_literal ();

public:
  inline literal_rep (symbol s2, int type2):
    s (s2), type (type2) {}

  friend class literal;
  friend symbol S (literal lit);
  friend ostream& operator << (ostream& out, literal s);
};

class literal {
DEFAULT_METHODS(literal);
public:
  inline literal (symbol s, int type):
    rep (new literal_rep (s, type)) {}

  inline bool operator == (literal lit) {
    return (rep->s == lit->s) && (rep->type == lit->type); }
  inline bool operator != (literal lit) {
    return (rep->s != lit->s) || (rep->type != lit->type); }
  inline friend symbol S (literal lit) { return lit->s; }
  friend ostream& operator << (ostream& out, literal s);
};

inline literal literal_int (symbol s) {
  return literal (s, LITERAL_INT_TYPE); }
inline literal literal_float (symbol s) {
  return literal (s, LITERAL_FLOAT_TYPE); }
inline literal literal_string (symbol s) {
  return literal (s, LITERAL_STRING_TYPE); }
inline string as_literal_contents (generic g) {
  return as_string (S (S (as_literal (g)))); }

}
#endif // __LITERAL_HH
