
/******************************************************************************
* MODULE     : symbol.hh
* DESCRIPTION: symbols
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __SYMBOL_HH
#define __SYMBOL_HH
#include <string.hh>
#include <table.hh>
namespace mmx {

class symbol_rep: public generic_rep {
protected:
  int      hash_value;
  generic  ref;

protected:
  size_t Size ();
  void Mark ();
  int Type ();
  int Hash ();
  bool Equal (generic x);
  generic Copy ();
  void Print (ostream& out);

public:
  symbol_rep (generic g);

  friend class symbol;
  friend symbol copy (symbol s);
  friend generic& S (symbol s);
  friend ostream& operator << (ostream& out, symbol s);
};

class symbol {
DEFAULT_METHODS(symbol);
public:
  symbol (char* s);
  symbol (string s);
  symbol (generic g);
  symbol (char* s, int begin, int end);
  symbol (string s, int begin, int end);
  symbol (generic g, int begin, int end);
  symbol (char* s, generic where);
  symbol (string s, generic where);
  symbol (generic g, generic where);

  inline bool operator == (symbol s) { return rep->ref.rep == s->ref.rep; }
  inline bool operator != (symbol s) { return rep->ref.rep != s->ref.rep; }
  inline friend symbol copy (symbol s) { return symbol (s.rep->ref); }
  inline friend generic& S (symbol s) { return s.rep->ref; }
  friend ostream& operator << (ostream& out, symbol s);
};

bool is_int (symbol s);
bool is_double (symbol s);
bool is_string (symbol s);
int as_int (symbol s);
double as_double (symbol s);
string as_string (symbol s);

inline bool
is_func (generic g, symbol s) {
  return is_list (g) && (N(g) > 0) && (car(g) == s);
}

inline bool
is_func (generic g, symbol s, int n) {
  return is_list (g) && (N(g) == (n+1)) && (car(g) == s);
}

}
#endif // __SYMBOL_HH
